import os
import torch

from typing import Literal, Any
from PIL.Image import Image
from .types import TableLayoutParsedFormat, GetModelDir
from .utils import expand_image


OutputFormat = Literal["latex", "markdown", "html"]

class Table:
  def __init__(
      self,
      device: Literal["cpu", "cuda"],
      get_model_dir: GetModelDir,
    ):
    self._model: Any | None = None
    self._model_path: str = get_model_dir()
    self._ban: bool = False
    if device == "cpu" or not torch.cuda.is_available():
      self._ban = True

  def predict(self, image: Image, format: TableLayoutParsedFormat) -> str | None:
    if self._ban:
      print("CUDA is not available. You cannot parse table from image.")
      return None

    output_format: str
    if format == TableLayoutParsedFormat.LATEX:
      output_format = "latex"
    elif format == TableLayoutParsedFormat.MARKDOWN:
      output_format = "markdown"
    elif format == TableLayoutParsedFormat.HTML:
      output_format = "html"
    else:
      raise ValueError(f"Table format {format} is not supported.")

    image = expand_image(image, 0.1)
    model = self._get_model()

    with torch.no_grad():
      results = model([image], output_format=output_format)

    if len(results) == 0:
      return None

    return results[0]

  def _get_model(self):
    if self._model is None:
      local_files_only: bool
      if os.path.exists(self._model_path):
        local_files_only = True
      else:
        local_files_only = False
        os.makedirs(self._model_path)

      from .struct_eqtable import build_model
      model = build_model(
        model_ckpt=self._model_path,
        max_new_tokens=1024,
        max_time=30,
        lmdeploy=False,
        flash_attn=True,
        batch_size=1,
        local_files_only=local_files_only,
      )
      self._model = model.cuda()
    return self._model


import torch
from torch import nn

from transformers import AutoTokenizer
try:
    from lmdeploy import pipeline, GenerationConfig, PytorchEngineConfig, ChatTemplateConfig
except:
    print("\033[93mimport lmdeploy failed, if do not use lmdeploy, ignore this message\033[0m")


class InternVL_LMDeploy(nn.Module):
    def __init__(
            self,
            model_path='U4R/StructTable-InternVL2-1B',
            max_new_tokens=1024,
            batch_size=4,
            cache_dir=None,
            local_files_only=None,
            **kwargs,
        ):
        super().__init__()
        self.model_path = model_path
        self.max_new_tokens = max_new_tokens
        self.max_batch_size = batch_size
        self.cache_dir = cache_dir
        self.local_files_only = local_files_only

        # init model and tokenizer from ckpt path
        self.init_tokenizer(model_path)
        self.init_model(model_path)

        self.prompt_template = {
            'latex': '<latex>',
            'html': '<html>',
            'markdown': '<markdown>',
        }
        # support output format
        self.supported_output_format = ['latex', 'html', 'markdown']

    def init_tokenizer(self, tokenizer_path):
        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=tokenizer_path,
            trust_remote_code=True,
            use_fast=False,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )

    def init_model(self, model_path):
        engine_config = PytorchEngineConfig(
            dtype='bfloat16',
            max_batch_size=self.max_batch_size,
            cache_max_entry_count=0.1
        )
        self.pipeline = pipeline(
            model_path,
            backend_config=engine_config,
            chat_template_config=ChatTemplateConfig(model_name='internvl2-internlm2')
        )

    def forward(self, images, output_format='latex', **kwargs):
        # process image to tokens
        if not isinstance(images, list):
            images = [images]

        prompts = [self.prompt_template[output_format]] * len(images)
        generation_config = GenerationConfig(
            max_new_tokens=self.max_new_tokens,
            do_sample=False,
            temperature=1.0,
            stop_token_ids=[self.tokenizer.eos_token_id],
        )

        responses = self.pipeline(
            [(x, y) for x, y in zip(prompts, images)],
            gen_config=generation_config,
        )
        batch_decode_texts = [responce.text for responce in responses]
        return batch_decode_texts





import os
import cv2
import copy

from . import predict_det
from . import predict_cls
from . import predict_rec
from .utils import get_rotate_crop_image, get_minarea_rect_crop

class TextSystem:
  def __init__(self, args):
    self.text_detector = predict_det.TextDetector(args)
    self.text_recognizer = predict_rec.TextRecognizer(args)
    self.use_angle_cls = True
    self.drop_score = args.drop_score
    if self.use_angle_cls:
      self.text_classifier = predict_cls.TextClassifier(args)

    self.args = args
    self.crop_image_res_index = 0

  def draw_crop_rec_res(self, output_dir, img_crop_list, rec_res):
    os.makedirs(output_dir, exist_ok=True)
    bbox_num = len(img_crop_list)
    for bno in range(bbox_num):
      cv2.imwrite(
        os.path.join(
          output_dir, f"mg_crop_{bno + self.crop_image_res_index}.jpg"
        ),
        img_crop_list[bno],
      )

    self.crop_image_res_index += bbox_num

  def __call__(self, img, cls=True):
    ori_im = img.copy()
    # 文字检测
    dt_boxes = self.text_detector(img)

    if dt_boxes is None:
      return None, None

    img_crop_list = []

    dt_boxes = sorted_boxes(dt_boxes)

    # 图片裁剪
    for bno in range(len(dt_boxes)):
      tmp_box = copy.deepcopy(dt_boxes[bno])
      if self.args.det_box_type == "quad":
        img_crop = get_rotate_crop_image(ori_im, tmp_box)
      else:
        img_crop = get_minarea_rect_crop(ori_im, tmp_box)
      img_crop_list.append(img_crop)

    # 方向分类
    if self.use_angle_cls and cls:
      img_crop_list, angle_list = self.text_classifier(img_crop_list)

    # 图像识别
    rec_res = self.text_recognizer(img_crop_list)

    if self.args.save_crop_res:
      self.draw_crop_rec_res(self.args.crop_res_save_dir, img_crop_list, rec_res)
    filter_boxes, filter_rec_res = [], []
    for box, rec_result in zip(dt_boxes, rec_res):
      text, score = rec_result
      if score >= self.drop_score:
        filter_boxes.append(box)
        filter_rec_res.append(rec_result)

    return filter_boxes, filter_rec_res


def sorted_boxes(dt_boxes):
  """
  Sort text boxes in order from top to bottom, left to right
  args:
    dt_boxes(array):detected text boxes with shape [4, 2]
  return:
    sorted boxes(array) with shape [4, 2]
  """
  num_boxes = dt_boxes.shape[0]
  sorted_boxes = sorted(dt_boxes, key=lambda x: (x[0][1], x[0][0]))
  _boxes = list(sorted_boxes)

  for i in range(num_boxes - 1):
    for j in range(i, -1, -1):
      if abs(_boxes[j + 1][0][1] - _boxes[j][0][1]) < 10 and (
        _boxes[j + 1][0][0] < _boxes[j][0][0]
      ):
        tmp = _boxes[j]
        _boxes[j] = _boxes[j + 1]
        _boxes[j + 1] = tmp
      else:
        break
  return _boxes



import numpy as np
import cv2
import os

from typing import Literal, Generator
from dataclasses import dataclass
from .onnxocr import TextSystem
from .types import GetModelDir, OCRFragment
from .rectangle import Rectangle
from .utils import is_space_text


_MODELS = (
  ("ppocrv4", "rec", "rec.onnx"),
  ("ppocrv4", "cls", "cls.onnx"),
  ("ppocrv4", "det", "det.onnx"),
  ("ch_ppocr_server_v2.0", "ppocr_keys_v1.txt"),
)

@dataclass
class _OONXParams:
  use_angle_cls: bool
  use_gpu: bool
  rec_image_shape: tuple[int, int, int]
  cls_image_shape: tuple[int, int, int]
  cls_batch_num: int
  cls_thresh: float
  label_list: list[str]

  det_algorithm: str
  det_limit_side_len: int
  det_limit_type: str
  det_db_thresh: float
  det_db_box_thresh: float
  det_db_unclip_ratio: float
  use_dilation: bool
  det_db_score_mode: str
  det_box_type: str
  rec_batch_num: int
  drop_score: float
  save_crop_res: bool
  rec_algorithm: str
  use_space_char: bool
  rec_model_dir: str
  cls_model_dir: str
  det_model_dir: str
  rec_char_dict_path: str




class OCR:
  def __init__(
      self,
      device: Literal["cpu", "cuda"],
      get_model_dir: GetModelDir,
    ):
    self._device: Literal["cpu", "cuda"] = device
    self._get_model_dir: GetModelDir = get_model_dir
    self._text_system: TextSystem | None = None

  def search_fragments(self, image: np.ndarray) -> Generator[OCRFragment, None, None]:
    for box, res in self._ocr(image):
      text, rank = res
      if is_space_text(text):
        continue

      rect = Rectangle(
        lt=(box[0][0], box[0][1]),
        rt=(box[1][0], box[1][1]),
        rb=(box[2][0], box[2][1]),
        lb=(box[3][0], box[3][1]),
      )
      if not rect.is_valid or rect.area == 0.0:
        continue

      yield OCRFragment(
        order=0,
        text=text,
        rank=rank,
        rect=rect,
      )

  def _ocr(self, image: np.ndarray) -> Generator[tuple[list[list[float]], tuple[str, float]], None, None]:
    text_system = self._get_text_system()
    image = self._preprocess_image(image)
    dt_boxes, rec_res = text_system(image)

    for box, res in zip(dt_boxes, rec_res):
      yield box.tolist(), res

  def make_model_paths(self) -> list[str]:
    model_paths = []
    model_dir = self._get_model_dir()
    for model_path in _MODELS:
      file_name = os.path.join(*model_path)
      model_paths.append(os.path.join(model_dir, file_name))
    return model_paths

  def _get_text_system(self) -> TextSystem:
    if self._text_system is None:
      model_paths = self.make_model_paths()
      self._text_system = TextSystem(_OONXParams(
        use_angle_cls=True,
        use_gpu=(self._device != "cpu"),
        rec_image_shape=(3, 48, 320),
        cls_image_shape=(3, 48, 192),
        cls_batch_num=6,
        cls_thresh=0.9,
        label_list=["0", "180"],
        det_algorithm="DB",
        det_limit_side_len=960,
        det_limit_type="max",
        det_db_thresh=0.3,
        det_db_box_thresh=0.6,
        det_db_unclip_ratio=1.5,
        use_dilation=False,
        det_db_score_mode="fast",
        det_box_type="quad",
        rec_batch_num=6,
        drop_score=0.5,
        save_crop_res=False,
        rec_algorithm="SVTR_LCNet",
        use_space_char=True,
        rec_model_dir=model_paths[0],
        cls_model_dir=model_paths[1],
        det_model_dir=model_paths[2],
        rec_char_dict_path=model_paths[3],
      ))

    return self._text_system

  def _preprocess_image(self, image: np.ndarray) -> np.ndarray:
    image = self._alpha_to_color(image, (255, 255, 255))
    # image = cv2.bitwise_not(image) # inv
    # image = self._binarize_img(image) # bin
    image = cv2.normalize(
      src=image,
      dst=np.zeros((image.shape[0], image.shape[1])),
      alpha=0,
      beta=255,
      norm_type=cv2.NORM_MINMAX,
    )
    image = cv2.fastNlMeansDenoisingColored(
      src=image,
      dst=None,
      h=10,
      hColor=10,
      templateWindowSize=7,
      searchWindowSize=15,
    )
    # image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY) # image to gray
    return image

  def _alpha_to_color(self, image: np.ndarray, alpha_color: tuple[float, float, float]) -> np.ndarray:
    if len(image.shape) == 3 and image.shape[2] == 4:
      B, G, R, A = cv2.split(image)
      alpha = A / 255

      R = (alpha_color[0] * (1 - alpha) + R * alpha).astype(np.uint8)
      G = (alpha_color[1] * (1 - alpha) + G * alpha).astype(np.uint8)
      B = (alpha_color[2] * (1 - alpha) + B * alpha).astype(np.uint8)

      image = cv2.merge((B, G, R))

    return image

  def _binarize_img(self, image: np.ndarray):
    if len(image.shape) == 3 and image.shape[2] == 3:
      gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)  # conversion to grayscale image
      # use cv2 threshold binarization
      _, gray = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
      image = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)
    return image



import numpy as np
import cv2
import sys
import math


class NormalizeImage(object):
  """ normalize image such as substract mean, divide std
  """

  def __init__(self, scale=None, mean=None, std=None, order='chw', **kwargs):
    if isinstance(scale, str):
      scale = eval(scale)
    self.scale = np.float32(scale if scale is not None else 1.0 / 255.0)
    mean = mean if mean is not None else [0.485, 0.456, 0.406]
    std = std if std is not None else [0.229, 0.224, 0.225]

    shape = (3, 1, 1) if order == 'chw' else (1, 1, 3)
    self.mean = np.array(mean).reshape(shape).astype('float32')
    self.std = np.array(std).reshape(shape).astype('float32')

  def __call__(self, data):
    img = data['image']
    from PIL import Image
    if isinstance(img, Image.Image):
      img = np.array(img)
    assert isinstance(img,
              np.ndarray), "invalid input 'img' in NormalizeImage"
    data['image'] = (
      img.astype('float32') * self.scale - self.mean) / self.std
    return data


class DetResizeForTest(object):
  def __init__(self, **kwargs):
    super(DetResizeForTest, self).__init__()
    self.resize_type = 0
    self.keep_ratio = False
    if 'image_shape' in kwargs:
      self.image_shape = kwargs['image_shape']
      self.resize_type = 1
      if 'keep_ratio' in kwargs:
        self.keep_ratio = kwargs['keep_ratio']
    elif 'limit_side_len' in kwargs:
      self.limit_side_len = kwargs['limit_side_len']
      self.limit_type = kwargs.get('limit_type', 'min')
    elif 'resize_long' in kwargs:
      self.resize_type = 2
      self.resize_long = kwargs.get('resize_long', 960)
    else:
      self.limit_side_len = 736
      self.limit_type = 'min'

  def __call__(self, data):
    img = data['image']
    src_h, src_w, _ = img.shape
    if sum([src_h, src_w]) < 64:
      img = self.image_padding(img)

    if self.resize_type == 0:
      # img, shape = self.resize_image_type0(img)
      img, [ratio_h, ratio_w] = self.resize_image_type0(img)
    elif self.resize_type == 2:
      img, [ratio_h, ratio_w] = self.resize_image_type2(img)
    else:
      # img, shape = self.resize_image_type1(img)
      img, [ratio_h, ratio_w] = self.resize_image_type1(img)
    data['image'] = img
    data['shape'] = np.array([src_h, src_w, ratio_h, ratio_w])
    return data

  def image_padding(self, im, value=0):
    h, w, c = im.shape
    im_pad = np.zeros((max(32, h), max(32, w), c), np.uint8) + value
    im_pad[:h, :w, :] = im
    return im_pad

  def resize_image_type1(self, img):
    resize_h, resize_w = self.image_shape
    ori_h, ori_w = img.shape[:2]  # (h, w, c)
    if self.keep_ratio is True:
      resize_w = ori_w * resize_h / ori_h
      N = math.ceil(resize_w / 32)
      resize_w = N * 32
    ratio_h = float(resize_h) / ori_h
    ratio_w = float(resize_w) / ori_w
    img = cv2.resize(img, (int(resize_w), int(resize_h)))
    # return img, np.array([ori_h, ori_w])
    return img, [ratio_h, ratio_w]

  def resize_image_type0(self, img):
    """
    resize image to a size multiple of 32 which is required by the network
    args:
      img(array): array with shape [h, w, c]
    return(tuple):
      img, (ratio_h, ratio_w)
    """
    limit_side_len = self.limit_side_len
    h, w, c = img.shape

    # limit the max side
    if self.limit_type == 'max':
      if max(h, w) > limit_side_len:
        if h > w:
          ratio = float(limit_side_len) / h
        else:
          ratio = float(limit_side_len) / w
      else:
        ratio = 1.
    elif self.limit_type == 'min':
      if min(h, w) < limit_side_len:
        if h < w:
          ratio = float(limit_side_len) / h
        else:
          ratio = float(limit_side_len) / w
      else:
        ratio = 1.
    elif self.limit_type == 'resize_long':
      ratio = float(limit_side_len) / max(h, w)
    else:
      raise Exception('not support limit type, image ')
    resize_h = int(h * ratio)
    resize_w = int(w * ratio)

    resize_h = max(int(round(resize_h / 32) * 32), 32)
    resize_w = max(int(round(resize_w / 32) * 32), 32)

    try:
      if int(resize_w) <= 0 or int(resize_h) <= 0:
        return None, (None, None)
      img = cv2.resize(img, (int(resize_w), int(resize_h)))
    except:
      print(img.shape, resize_w, resize_h)
      sys.exit(0)
    ratio_h = resize_h / float(h)
    ratio_w = resize_w / float(w)
    return img, [ratio_h, ratio_w]

  def resize_image_type2(self, img):
    h, w, _ = img.shape

    resize_w = w
    resize_h = h

    if resize_h > resize_w:
      ratio = float(self.resize_long) / resize_h
    else:
      ratio = float(self.resize_long) / resize_w

    resize_h = int(resize_h * ratio)
    resize_w = int(resize_w * ratio)

    max_stride = 128
    resize_h = (resize_h + max_stride - 1) // max_stride * max_stride
    resize_w = (resize_w + max_stride - 1) // max_stride * max_stride
    img = cv2.resize(img, (int(resize_w), int(resize_h)))
    ratio_h = resize_h / float(h)
    ratio_w = resize_w / float(w)

    return img, [ratio_h, ratio_w]

class ToCHWImage(object):
  """ convert hwc image to chw image
  """

  def __init__(self, **kwargs):
    pass

  def __call__(self, data):
    img = data['image']
    from PIL import Image
    if isinstance(img, Image.Image):
      img = np.array(img)
    data['image'] = img.transpose((2, 0, 1))
    return data


class KeepKeys(object):
  def __init__(self, keep_keys, **kwargs):
    self.keep_keys = keep_keys

  def __call__(self, data):
    data_list = []
    for key in self.keep_keys:
      data_list.append(data[key])
    return data_list


class ClsPostProcess  (object):
  """ Convert between text-label and text-index """

  def __init__(self, label_list=None, key=None, **kwargs):
    super(ClsPostProcess, self).__init__()
    self.label_list = label_list
    self.key = key

  def __call__(self, preds, label=None, *args, **kwargs):
    if self.key is not None:
      preds = preds[self.key]

    label_list = self.label_list
    if label_list is None:
      label_list = {idx: idx for idx in range(preds.shape[-1])}

    # if isinstance(preds, paddle.Tensor):
    #   preds = preds.numpy()

    pred_idxs = preds.argmax(axis=1)
    decode_out = [(label_list[idx], preds[i, idx])
            for i, idx in enumerate(pred_idxs)]
    if label is None:
      return decode_out
    label = [(label_list[idx], 1.0) for idx in label]
    return decode_out, label



from math import pi, atan2, sqrt, sin, cos
from .types import OCRFragment
from .rectangle import Point, Rectangle


class RotationAdjuster:
  def __init__(
      self,
      origin_size: tuple[int, int],
      new_size: tuple[int, int],
      rotation: float,
      to_origin_coordinate: bool,
    ):
    from_size: tuple[int, int]
    to_size: tuple[int, int]
    if to_origin_coordinate:
      from_size = new_size
      to_size = origin_size
    else:
      from_size = origin_size
      to_size = new_size
      rotation = -rotation

    self._rotation: float = rotation
    self._center_offset: tuple[float, float] = (
      - from_size[0] / 2.0,
      - from_size[1] / 2.0,
    )
    self._new_offset: tuple[float, float] = (
      to_size[0] / 2.0,
      to_size[1] / 2.0,
    )

  def adjust(self, point: Point) -> Point:
    x, y = point
    x += self._center_offset[0]
    y += self._center_offset[1]

    if x != 0.0 or y != 0.0:
      radius = sqrt(x*x + y*y)
      angle = atan2(y, x) + self._rotation
      x = radius * cos(angle)
      y = radius * sin(angle)

    x += self._new_offset[0]
    y += self._new_offset[1]

    return x, y

# to [0, pi)
def normal_vertical_rotation(rotation: float) -> float:
  while rotation >= 2 * pi:
    rotation -= 2 * pi
  while rotation <= - 2 * pi:
    rotation += 2 * pi
  if rotation < 0.0:
    rotation += pi
  return rotation

def calculate_rotation(fragments: list[OCRFragment]):
  horizontal_rotations: list[float] = []
  vertical_rotations: list[float] = []

  for fragment in fragments:
    result = _rotation_with(fragment.rect)
    if result is not None:
      horizontal_rotations.extend(result[0])
      vertical_rotations.extend(result[1])

  if len(horizontal_rotations) == 0 or len(vertical_rotations) == 0:
    return 0.0

  horizontal_rotations = _normal_horizontal_rotations(horizontal_rotations)
  horizontal_rotation = _find_median(horizontal_rotations)
  vertical_rotation = _find_median(vertical_rotations)

  return (vertical_rotation - 0.5 * pi + horizontal_rotation) / 2.0

# @return horizontal [-pi/2, pi/2), vertical [0, pi)
def calculate_rotation_with_rect(rect: Rectangle) -> tuple[float, float]:
  result = _rotation_with(rect)
  if result is None:
    return 0.0, 0.5 * pi

  horizontal_rotations, vertical_rotations = result
  horizontal_rotations = _normal_horizontal_rotations(horizontal_rotations)
  horizontal_rotation = _find_mean(horizontal_rotations)
  vertical_rotation = _find_mean(vertical_rotations)

  return horizontal_rotation, vertical_rotation

def _rotation_with(rect: Rectangle):
  rotations0: list[float] = []
  rotations1: list[float] = []

  for i, (p1, p2) in enumerate(rect.segments):
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    if dx == 0.0 and dy == 0.0:
      return None
    rotation: float = atan2(dy, dx)
    if rotation < 0.0:
      rotation += pi
    if i % 2 == 0:
      rotations0.append(rotation)
    else:
      rotations1.append(rotation)

  if _is_vertical(rotations0):
    return rotations1, rotations0
  else:
    return rotations0, rotations1

# [0, pi) --> [-pi/2, pi/2)
def _normal_horizontal_rotations(rotations: list[float]) -> list[float]:
  for i, rotation in enumerate(rotations):
    if rotation > 0.5 * pi:
      rotations[i] = rotation - pi
  return rotations

def _find_median(rotations: list[float]):
  rotations.sort()
  n = len(rotations)

  if n % 2 == 1:
    return rotations[n // 2]
  else:
    mid1 = rotations[n // 2 - 1]
    mid2 = rotations[n // 2]
    return (mid1 + mid2) / 2

def _find_mean(rotations: list[float]) -> float:
  if len(rotations) == 0:
    return 0.0
  return sum(rotations) / len(rotations)

# rotation is in [0, pi)
def _is_vertical(rotations: list[float]):
  horizontal_count: int = 0
  vertical_count: int = 0
  horizontal_delta: float = 0.0
  vertical_delta: float = 0.0

  for rotation in rotations:
    if rotation < 0.25 * pi: # [0, pi/4)
      horizontal_count += 1
      horizontal_delta += rotation
    elif rotation < 0.75 * pi: # [pi/4, 3pi/4)
      vertical_count += 1
      vertical_delta += abs(rotation - 0.5 * pi)
    else: # [3pi/4, pi)
      horizontal_count += 1
      horizontal_delta += pi - rotation

  if vertical_count == horizontal_delta:
    return vertical_delta < horizontal_delta
  else:
    return vertical_count > horizontal_count



import re
import numpy as np

paddle = None


class BaseRecLabelDecode(object):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False):
    self.beg_str = "sos"
    self.end_str = "eos"
    self.reverse = False
    self.character_str = []

    if character_dict_path is None:
      self.character_str = "0123456789abcdefghijklmnopqrstuvwxyz"
      dict_character = list(self.character_str)
    else:
      with open(character_dict_path, "rb") as fin:
        lines = fin.readlines()
        for line in lines:
          line = line.decode("utf-8").strip("\n").strip("\r\n")
          self.character_str.append(line)
      if use_space_char:
        self.character_str.append(" ")
      dict_character = list(self.character_str)
      if "arabic" in character_dict_path:
        self.reverse = True

    dict_character = self.add_special_char(dict_character)
    self.dict = {}
    for i, char in enumerate(dict_character):
      self.dict[char] = i
    self.character = dict_character

  def pred_reverse(self, pred):
    pred_re = []
    c_current = ""
    for c in pred:
      if not bool(re.search("[a-zA-Z0-9 :*./%+-]", c)):
        if c_current != "":
          pred_re.append(c_current)
        pred_re.append(c)
        c_current = ""
      else:
        c_current += c
    if c_current != "":
      pred_re.append(c_current)

    return "".join(pred_re[::-1])

  def add_special_char(self, dict_character):
    return dict_character

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    ignored_tokens = self.get_ignored_tokens()
    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      selection = np.ones(len(text_index[batch_idx]), dtype=bool)
      if is_remove_duplicate:
        selection[1:] = text_index[batch_idx][1:] != text_index[batch_idx][:-1]
      for ignored_token in ignored_tokens:
        selection &= text_index[batch_idx] != ignored_token

      char_list = [
        self.character[text_id] for text_id in text_index[batch_idx][selection]
      ]
      if text_prob is not None:
        conf_list = text_prob[batch_idx][selection]
      else:
        conf_list = [1] * len(selection)
      if len(conf_list) == 0:
        conf_list = [0]

      text = "".join(char_list)

      if self.reverse:  # for arabic rec
        text = self.pred_reverse(text)

      result_list.append((text, np.mean(conf_list).tolist()))
    return result_list

  def get_ignored_tokens(self):
    return [0]  # for ctc blank


class CTCLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(CTCLabelDecode, self).__init__(character_dict_path, use_space_char)

  def __call__(self, preds, label=None, *args, **kwargs):
    if isinstance(preds, tuple) or isinstance(preds, list):
      preds = preds[-1]
    # if isinstance(preds, paddle.Tensor):
    #   preds = preds.numpy()
    preds_idx = preds.argmax(axis=2)
    preds_prob = preds.max(axis=2)
    text = self.decode(preds_idx, preds_prob, is_remove_duplicate=True)
    if label is None:
      return text
    label = self.decode(label)
    return text, label

  def add_special_char(self, dict_character):
    dict_character = ["blank"] + dict_character
    return dict_character


class DistillationCTCLabelDecode(CTCLabelDecode):
  """
  Convert
  Convert between text-label and text-index
  """

  def __init__(
    self,
    character_dict_path=None,
    use_space_char=False,
    model_name=["student"],
    key=None,
    multi_head=False,
    **kwargs
  ):
    super(DistillationCTCLabelDecode, self).__init__(
      character_dict_path, use_space_char
    )
    if not isinstance(model_name, list):
      model_name = [model_name]
    self.model_name = model_name

    self.key = key
    self.multi_head = multi_head

  def __call__(self, preds, label=None, *args, **kwargs):
    output = dict()
    for name in self.model_name:
      pred = preds[name]
      if self.key is not None:
        pred = pred[self.key]
      if self.multi_head and isinstance(pred, dict):
        pred = pred["ctc"]
      output[name] = super().__call__(pred, label=label, *args, **kwargs)
    return output


class AttnLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(AttnLabelDecode, self).__init__(character_dict_path, use_space_char)

  def add_special_char(self, dict_character):
    self.beg_str = "sos"
    self.end_str = "eos"
    dict_character = dict_character
    dict_character = [self.beg_str] + dict_character + [self.end_str]
    return dict_character

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    ignored_tokens = self.get_ignored_tokens()
    [beg_idx, end_idx] = self.get_ignored_tokens()
    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        if text_index[batch_idx][idx] in ignored_tokens:
          continue
        if int(text_index[batch_idx][idx]) == int(end_idx):
          break
        if is_remove_duplicate:
          # only for predict
          if (
            idx > 0
            and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]
          ):
            continue
        char_list.append(self.character[int(text_index[batch_idx][idx])])
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)
      text = "".join(char_list)
      result_list.append((text, np.mean(conf_list).tolist()))
    return result_list

  def __call__(self, preds, label=None, *args, **kwargs):
    """
    text = self.decode(text)
    if label is None:
      return text
    else:
      label = self.decode(label, is_remove_duplicate=False)
      return text, label
    """
    if isinstance(preds, paddle.Tensor):
      preds = preds.numpy()

    preds_idx = preds.argmax(axis=2)
    preds_prob = preds.max(axis=2)
    text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
    if label is None:
      return text
    label = self.decode(label, is_remove_duplicate=False)
    return text, label

  def get_ignored_tokens(self):
    beg_idx = self.get_beg_end_flag_idx("beg")
    end_idx = self.get_beg_end_flag_idx("end")
    return [beg_idx, end_idx]

  def get_beg_end_flag_idx(self, beg_or_end):
    if beg_or_end == "beg":
      idx = np.array(self.dict[self.beg_str])
    elif beg_or_end == "end":
      idx = np.array(self.dict[self.end_str])
    else:
      assert False, "unsupport type %s in get_beg_end_flag_idx" % beg_or_end
    return idx


class RFLLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(RFLLabelDecode, self).__init__(character_dict_path, use_space_char)

  def add_special_char(self, dict_character):
    self.beg_str = "sos"
    self.end_str = "eos"
    dict_character = dict_character
    dict_character = [self.beg_str] + dict_character + [self.end_str]
    return dict_character

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    ignored_tokens = self.get_ignored_tokens()
    [beg_idx, end_idx] = self.get_ignored_tokens()
    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        if text_index[batch_idx][idx] in ignored_tokens:
          continue
        if int(text_index[batch_idx][idx]) == int(end_idx):
          break
        if is_remove_duplicate:
          # only for predict
          if (
            idx > 0
            and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]
          ):
            continue
        char_list.append(self.character[int(text_index[batch_idx][idx])])
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)
      text = "".join(char_list)
      result_list.append((text, np.mean(conf_list).tolist()))
    return result_list

  def __call__(self, preds, label=None, *args, **kwargs):
    # if seq_outputs is not None:
    if isinstance(preds, tuple) or isinstance(preds, list):
      cnt_outputs, seq_outputs = preds
      if isinstance(seq_outputs, paddle.Tensor):
        seq_outputs = seq_outputs.numpy()
      preds_idx = seq_outputs.argmax(axis=2)
      preds_prob = seq_outputs.max(axis=2)
      text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)

      if label is None:
        return text
      label = self.decode(label, is_remove_duplicate=False)
      return text, label

    else:
      cnt_outputs = preds
      if isinstance(cnt_outputs, paddle.Tensor):
        cnt_outputs = cnt_outputs.numpy()
      cnt_length = []
      for lens in cnt_outputs:
        length = round(np.sum(lens))
        cnt_length.append(length)
      if label is None:
        return cnt_length
      label = self.decode(label, is_remove_duplicate=False)
      length = [len(res[0]) for res in label]
      return cnt_length, length

  def get_ignored_tokens(self):
    beg_idx = self.get_beg_end_flag_idx("beg")
    end_idx = self.get_beg_end_flag_idx("end")
    return [beg_idx, end_idx]

  def get_beg_end_flag_idx(self, beg_or_end):
    if beg_or_end == "beg":
      idx = np.array(self.dict[self.beg_str])
    elif beg_or_end == "end":
      idx = np.array(self.dict[self.end_str])
    else:
      assert False, "unsupport type %s in get_beg_end_flag_idx" % beg_or_end
    return idx


class SEEDLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(SEEDLabelDecode, self).__init__(character_dict_path, use_space_char)

  def add_special_char(self, dict_character):
    self.padding_str = "padding"
    self.end_str = "eos"
    self.unknown = "unknown"
    dict_character = dict_character + [self.end_str, self.padding_str, self.unknown]
    return dict_character

  def get_ignored_tokens(self):
    end_idx = self.get_beg_end_flag_idx("eos")
    return [end_idx]

  def get_beg_end_flag_idx(self, beg_or_end):
    if beg_or_end == "sos":
      idx = np.array(self.dict[self.beg_str])
    elif beg_or_end == "eos":
      idx = np.array(self.dict[self.end_str])
    else:
      assert False, "unsupport type %s in get_beg_end_flag_idx" % beg_or_end
    return idx

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    [end_idx] = self.get_ignored_tokens()
    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        if int(text_index[batch_idx][idx]) == int(end_idx):
          break
        if is_remove_duplicate:
          # only for predict
          if (
            idx > 0
            and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]
          ):
            continue
        char_list.append(self.character[int(text_index[batch_idx][idx])])
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)
      text = "".join(char_list)
      result_list.append((text, np.mean(conf_list).tolist()))
    return result_list

  def __call__(self, preds, label=None, *args, **kwargs):
    """
    text = self.decode(text)
    if label is None:
      return text
    else:
      label = self.decode(label, is_remove_duplicate=False)
      return text, label
    """
    preds_idx = preds["rec_pred"]
    if isinstance(preds_idx, paddle.Tensor):
      preds_idx = preds_idx.numpy()
    if "rec_pred_scores" in preds:
      preds_idx = preds["rec_pred"]
      preds_prob = preds["rec_pred_scores"]
    else:
      preds_idx = preds["rec_pred"].argmax(axis=2)
      preds_prob = preds["rec_pred"].max(axis=2)
    text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
    if label is None:
      return text
    label = self.decode(label, is_remove_duplicate=False)
    return text, label


class SRNLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(SRNLabelDecode, self).__init__(character_dict_path, use_space_char)
    self.max_text_length = kwargs.get("max_text_length", 25)

  def __call__(self, preds, label=None, *args, **kwargs):
    pred = preds["predict"]
    char_num = len(self.character_str) + 2
    if isinstance(pred, paddle.Tensor):
      pred = pred.numpy()
    pred = np.reshape(pred, [-1, char_num])

    preds_idx = np.argmax(pred, axis=1)
    preds_prob = np.max(pred, axis=1)

    preds_idx = np.reshape(preds_idx, [-1, self.max_text_length])

    preds_prob = np.reshape(preds_prob, [-1, self.max_text_length])

    text = self.decode(preds_idx, preds_prob)

    if label is None:
      text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
      return text
    label = self.decode(label)
    return text, label

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    ignored_tokens = self.get_ignored_tokens()
    batch_size = len(text_index)

    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        if text_index[batch_idx][idx] in ignored_tokens:
          continue
        if is_remove_duplicate:
          # only for predict
          if (
            idx > 0
            and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]
          ):
            continue
        char_list.append(self.character[int(text_index[batch_idx][idx])])
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)

      text = "".join(char_list)
      result_list.append((text, np.mean(conf_list).tolist()))
    return result_list

  def add_special_char(self, dict_character):
    dict_character = dict_character + [self.beg_str, self.end_str]
    return dict_character

  def get_ignored_tokens(self):
    beg_idx = self.get_beg_end_flag_idx("beg")
    end_idx = self.get_beg_end_flag_idx("end")
    return [beg_idx, end_idx]

  def get_beg_end_flag_idx(self, beg_or_end):
    if beg_or_end == "beg":
      idx = np.array(self.dict[self.beg_str])
    elif beg_or_end == "end":
      idx = np.array(self.dict[self.end_str])
    else:
      assert False, "unsupport type %s in get_beg_end_flag_idx" % beg_or_end
    return idx


class SARLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(SARLabelDecode, self).__init__(character_dict_path, use_space_char)

    self.rm_symbol = kwargs.get("rm_symbol", False)

  def add_special_char(self, dict_character):
    beg_end_str = "<BOS/EOS>"
    unknown_str = "<UKN>"
    padding_str = "<PAD>"
    dict_character = dict_character + [unknown_str]
    self.unknown_idx = len(dict_character) - 1
    dict_character = dict_character + [beg_end_str]
    self.start_idx = len(dict_character) - 1
    self.end_idx = len(dict_character) - 1
    dict_character = dict_character + [padding_str]
    self.padding_idx = len(dict_character) - 1
    return dict_character

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    ignored_tokens = self.get_ignored_tokens()

    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        if text_index[batch_idx][idx] in ignored_tokens:
          continue
        if int(text_index[batch_idx][idx]) == int(self.end_idx):
          if text_prob is None and idx == 0:
            continue
          else:
            break
        if is_remove_duplicate:
          # only for predict
          if (
            idx > 0
            and text_index[batch_idx][idx - 1] == text_index[batch_idx][idx]
          ):
            continue
        char_list.append(self.character[int(text_index[batch_idx][idx])])
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)
      text = "".join(char_list)
      if self.rm_symbol:
        comp = re.compile("[^A-Z^a-z^0-9^\u4e00-\u9fa5]")
        text = text.lower()
        text = comp.sub("", text)
      result_list.append((text, np.mean(conf_list).tolist()))
    return result_list

  def __call__(self, preds, label=None, *args, **kwargs):
    if isinstance(preds, paddle.Tensor):
      preds = preds.numpy()
    preds_idx = preds.argmax(axis=2)
    preds_prob = preds.max(axis=2)

    text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)

    if label is None:
      return text
    label = self.decode(label, is_remove_duplicate=False)
    return text, label

  def get_ignored_tokens(self):
    return [self.padding_idx]


class DistillationSARLabelDecode(SARLabelDecode):
  """
  Convert
  Convert between text-label and text-index
  """

  def __init__(
    self,
    character_dict_path=None,
    use_space_char=False,
    model_name=["student"],
    key=None,
    multi_head=False,
    **kwargs
  ):
    super(DistillationSARLabelDecode, self).__init__(
      character_dict_path, use_space_char
    )
    if not isinstance(model_name, list):
      model_name = [model_name]
    self.model_name = model_name

    self.key = key
    self.multi_head = multi_head

  def __call__(self, preds, label=None, *args, **kwargs):
    output = dict()
    for name in self.model_name:
      pred = preds[name]
      if self.key is not None:
        pred = pred[self.key]
      if self.multi_head and isinstance(pred, dict):
        pred = pred["sar"]
      output[name] = super().__call__(pred, label=label, *args, **kwargs)
    return output


class PRENLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(PRENLabelDecode, self).__init__(character_dict_path, use_space_char)

  def add_special_char(self, dict_character):
    padding_str = "<PAD>"  # 0
    end_str = "<EOS>"  # 1
    unknown_str = "<UNK>"  # 2

    dict_character = [padding_str, end_str, unknown_str] + dict_character
    self.padding_idx = 0
    self.end_idx = 1
    self.unknown_idx = 2

    return dict_character

  def decode(self, text_index, text_prob=None):
    """convert text-index into text-label."""
    result_list = []
    batch_size = len(text_index)

    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        if text_index[batch_idx][idx] == self.end_idx:
          break
        if text_index[batch_idx][idx] in [self.padding_idx, self.unknown_idx]:
          continue
        char_list.append(self.character[int(text_index[batch_idx][idx])])
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)

      text = "".join(char_list)
      if len(text) > 0:
        result_list.append((text, np.mean(conf_list).tolist()))
      else:
        # here confidence of empty recog result is 1
        result_list.append(("", 1))
    return result_list

  def __call__(self, preds, label=None, *args, **kwargs):
    if isinstance(preds, paddle.Tensor):
      preds = preds.numpy()
    preds_idx = preds.argmax(axis=2)
    preds_prob = preds.max(axis=2)
    text = self.decode(preds_idx, preds_prob)
    if label is None:
      return text
    label = self.decode(label)
    return text, label


class NRTRLabelDecode(BaseRecLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=True, **kwargs):
    super(NRTRLabelDecode, self).__init__(character_dict_path, use_space_char)

  def __call__(self, preds, label=None, *args, **kwargs):

    if len(preds) == 2:
      preds_id = preds[0]
      preds_prob = preds[1]
      if isinstance(preds_id, paddle.Tensor):
        preds_id = preds_id.numpy()
      if isinstance(preds_prob, paddle.Tensor):
        preds_prob = preds_prob.numpy()
      if preds_id[0][0] == 2:
        preds_idx = preds_id[:, 1:]
        preds_prob = preds_prob[:, 1:]
      else:
        preds_idx = preds_id
      text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
      if label is None:
        return text
      label = self.decode(label[:, 1:])
    else:
      if isinstance(preds, paddle.Tensor):
        preds = preds.numpy()
      preds_idx = preds.argmax(axis=2)
      preds_prob = preds.max(axis=2)
      text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
      if label is None:
        return text
      label = self.decode(label[:, 1:])
    return text, label

  def add_special_char(self, dict_character):
    dict_character = ["blank", "<unk>", "<s>", "</s>"] + dict_character
    return dict_character

  def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
    """convert text-index into text-label."""
    result_list = []
    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      char_list = []
      conf_list = []
      for idx in range(len(text_index[batch_idx])):
        try:
          char_idx = self.character[int(text_index[batch_idx][idx])]
        except:
          continue
        if char_idx == "</s>":  # end
          break
        char_list.append(char_idx)
        if text_prob is not None:
          conf_list.append(text_prob[batch_idx][idx])
        else:
          conf_list.append(1)
      text = "".join(char_list)
      result_list.append((text.lower(), np.mean(conf_list).tolist()))
    return result_list


class ViTSTRLabelDecode(NRTRLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(ViTSTRLabelDecode, self).__init__(character_dict_path, use_space_char)

  def __call__(self, preds, label=None, *args, **kwargs):
    if isinstance(preds, paddle.Tensor):
      preds = preds[:, 1:].numpy()
    else:
      preds = preds[:, 1:]
    preds_idx = preds.argmax(axis=2)
    preds_prob = preds.max(axis=2)
    text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
    if label is None:
      return text
    label = self.decode(label[:, 1:])
    return text, label

  def add_special_char(self, dict_character):
    dict_character = ["<s>", "</s>"] + dict_character
    return dict_character


class ABINetLabelDecode(NRTRLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(ABINetLabelDecode, self).__init__(character_dict_path, use_space_char)

  def __call__(self, preds, label=None, *args, **kwargs):
    if isinstance(preds, dict):
      preds = preds["align"][-1].numpy()
    elif isinstance(preds, paddle.Tensor):
      preds = preds.numpy()
    else:
      preds = preds

    preds_idx = preds.argmax(axis=2)
    preds_prob = preds.max(axis=2)
    text = self.decode(preds_idx, preds_prob, is_remove_duplicate=False)
    if label is None:
      return text
    label = self.decode(label)
    return text, label

  def add_special_char(self, dict_character):
    dict_character = ["</s>"] + dict_character
    return dict_character


class SPINLabelDecode(AttnLabelDecode):
  """Convert between text-label and text-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(SPINLabelDecode, self).__init__(character_dict_path, use_space_char)

  def add_special_char(self, dict_character):
    self.beg_str = "sos"
    self.end_str = "eos"
    dict_character = dict_character
    dict_character = [self.beg_str] + [self.end_str] + dict_character
    return dict_character


# class VLLabelDecode(BaseRecLabelDecode):
#   """ Convert between text-label and text-index """
#
#   def __init__(self, character_dict_path=None, use_space_char=False,
#          **kwargs):
#     super(VLLabelDecode, self).__init__(character_dict_path, use_space_char)
#     self.max_text_length = kwargs.get('max_text_length', 25)
#     self.nclass = len(self.character) + 1
#
#   def decode(self, text_index, text_prob=None, is_remove_duplicate=False):
#     """ convert text-index into text-label. """
#     result_list = []
#     ignored_tokens = self.get_ignored_tokens()
#     batch_size = len(text_index)
#     for batch_idx in range(batch_size):
#       selection = np.ones(len(text_index[batch_idx]), dtype=bool)
#       if is_remove_duplicate:
#         selection[1:] = text_index[batch_idx][1:] != text_index[
#           batch_idx][:-1]
#       for ignored_token in ignored_tokens:
#         selection &= text_index[batch_idx] != ignored_token
#
#       char_list = [
#         self.character[text_id - 1]
#         for text_id in text_index[batch_idx][selection]
#       ]
#       if text_prob is not None:
#         conf_list = text_prob[batch_idx][selection]
#       else:
#         conf_list = [1] * len(selection)
#       if len(conf_list) == 0:
#         conf_list = [0]
#
#       text = ''.join(char_list)
#       result_list.append((text, np.mean(conf_list).tolist()))
#     return result_list
#
#   def __call__(self, preds, label=None, length=None, *args, **kwargs):
#     if len(preds) == 2:  # eval mode
#       text_pre, x = preds
#       b = text_pre.shape[1]
#       lenText = self.max_text_length
#       nsteps = self.max_text_length
#
#       if not isinstance(text_pre, paddle.Tensor):
#         text_pre = paddle.to_tensor(text_pre, dtype='float32')
#
#       out_res = paddle.zeros(
#         shape=[lenText, b, self.nclass], dtype=x.dtype)
#       out_length = paddle.zeros(shape=[b], dtype=x.dtype)
#       now_step = 0
#       for _ in range(nsteps):
#         if 0 in out_length and now_step < nsteps:
#           tmp_result = text_pre[now_step, :, :]
#           out_res[now_step] = tmp_result
#           tmp_result = tmp_result.topk(1)[1].squeeze(axis=1)
#           for j in range(b):
#             if out_length[j] == 0 and tmp_result[j] == 0:
#               out_length[j] = now_step + 1
#           now_step += 1
#       for j in range(0, b):
#         if int(out_length[j]) == 0:
#           out_length[j] = nsteps
#       start = 0
#       output = paddle.zeros(
#         shape=[int(out_length.sum()), self.nclass], dtype=x.dtype)
#       for i in range(0, b):
#         cur_length = int(out_length[i])
#         output[start:start + cur_length] = out_res[0:cur_length, i, :]
#         start += cur_length
#       net_out = output
#       length = out_length
#
#     else:  # train mode
#       net_out = preds[0]
#       length = length
#       net_out = paddle.concat([t[:l] for t, l in zip(net_out, length)])
#     text = []
#     if not isinstance(net_out, paddle.Tensor):
#       net_out = paddle.to_tensor(net_out, dtype='float32')
#     net_out = F.softmax(net_out, axis=1)
#     for i in range(0, length.shape[0]):
#       preds_idx = net_out[int(length[:i].sum()):int(length[:i].sum(
#       ) + length[i])].topk(1)[1][:, 0].tolist()
#       preds_text = ''.join([
#         self.character[idx - 1]
#         if idx > 0 and idx <= len(self.character) else ''
#         for idx in preds_idx
#       ])
#       preds_prob = net_out[int(length[:i].sum()):int(length[:i].sum(
#       ) + length[i])].topk(1)[0][:, 0]
#       preds_prob = paddle.exp(
#         paddle.log(preds_prob).sum() / (preds_prob.shape[0] + 1e-6))
#       text.append((preds_text, preds_prob.numpy()[0]))
#     if label is None:
#       return text
#     label = self.decode(label)
#     return text, label


class CANLabelDecode(BaseRecLabelDecode):
  """Convert between latex-symbol and symbol-index"""

  def __init__(self, character_dict_path=None, use_space_char=False, **kwargs):
    super(CANLabelDecode, self).__init__(character_dict_path, use_space_char)

  def decode(self, text_index, preds_prob=None):
    result_list = []
    batch_size = len(text_index)
    for batch_idx in range(batch_size):
      seq_end = text_index[batch_idx].argmin(0)
      idx_list = text_index[batch_idx][:seq_end].tolist()
      symbol_list = [self.character[idx] for idx in idx_list]
      probs = []
      if preds_prob is not None:
        probs = preds_prob[batch_idx][: len(symbol_list)].tolist()

      result_list.append([" ".join(symbol_list), probs])
    return result_list

  def __call__(self, preds, label=None, *args, **kwargs):
    pred_prob, _, _, _ = preds
    preds_idx = pred_prob.argmax(axis=2)

    text = self.decode(preds_idx)
    if label is None:
      return text
    label = self.decode(label)
    return text, label



from dataclasses import dataclass
from typing import Literal, Callable, Protocol, runtime_checkable, List
from enum import auto, Enum
from PIL.Image import Image
from .rectangle import Rectangle


@dataclass
class OCRFragment:
  order: int
  text: str
  rank: float
  rect: Rectangle

class LayoutClass(Enum):
  TITLE = 0
  PLAIN_TEXT = 1
  ABANDON = 2
  FIGURE = 3
  FIGURE_CAPTION = 4
  TABLE = 5
  TABLE_CAPTION = 6
  TABLE_FOOTNOTE = 7
  ISOLATE_FORMULA = 8
  FORMULA_CAPTION = 9

class TableLayoutParsedFormat(Enum):
  LATEX = auto()
  MARKDOWN = auto()
  HTML = auto()

@dataclass
class BaseLayout:
  rect: Rectangle
  fragments: List[OCRFragment]

@dataclass
class PlainLayout(BaseLayout):
  cls: Literal[
    LayoutClass.TITLE,
    LayoutClass.PLAIN_TEXT,
    LayoutClass.ABANDON,
    LayoutClass.FIGURE,
    LayoutClass.FIGURE_CAPTION,
    LayoutClass.TABLE_CAPTION,
    LayoutClass.TABLE_FOOTNOTE,
    LayoutClass.FORMULA_CAPTION,
  ]

@dataclass
class TableLayout(BaseLayout):
  parsed: tuple[str, TableLayoutParsedFormat] | None
  cls: LayoutClass.TABLE

@dataclass
class FormulaLayout(BaseLayout):
  latex: str | None
  cls: LayoutClass.ISOLATE_FORMULA

Layout = PlainLayout | TableLayout | FormulaLayout

@dataclass
class ExtractedResult:
  rotation: float
  layouts: List[Layout]
  extracted_image: Image
  adjusted_image: Image | None

GetModelDir = Callable[[], str]


@runtime_checkable
class ModelsDownloader(Protocol):

  def onnx_ocr(self) -> str:
    pass

  def yolo(self) -> str:
    pass

  def layoutreader(self) -> str:
    pass

  def struct_eqtable(self) -> str:
    pass

  def latex(self) -> str:
    pass



import torch

from typing import Generator
from dataclasses import dataclass
from transformers import LayoutLMv3ForTokenClassification

from .types import Layout, LayoutClass, GetModelDir
from .layoutreader import prepare_inputs, boxes2inputs, parse_logits


@dataclass
class _BBox:
  layout_index: int
  fragment_index: int
  virtual: bool
  order: int
  value: tuple[float, float, float, float]

class LayoutOrder:
  def __init__(self, get_model_dir: GetModelDir):
    self._model_path: str = get_model_dir()
    self._model: LayoutLMv3ForTokenClassification | None = None
    self._device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

  def _get_model(self) -> LayoutLMv3ForTokenClassification:
    if self._model is None:
      self._model = LayoutLMv3ForTokenClassification.from_pretrained(
        pretrained_model_name_or_path=self._model_path,
        local_files_only=True,
      ).to(device=self._device)
    return self._model

  def sort(self, layouts: list[Layout], size: tuple[int, int]) -> list[Layout]:
    width, height = size
    if width == 0 or height == 0:
      return layouts

    bbox_list = self._order_and_get_bbox_list(
      layouts=layouts,
      width=width,
      height=height,
    )
    if bbox_list is None:
      return layouts

    return self._sort_layouts_and_fragments(layouts, bbox_list)

  def _order_and_get_bbox_list(
      self,
      layouts: list[Layout],
      width: int,
      height: int,
    ) -> list[_BBox] | None:

    line_height = self._line_height(layouts)
    bbox_list: list[_BBox] = []

    for i, layout in enumerate(layouts):
      if layout.cls == LayoutClass.PLAIN_TEXT and \
         len(layout.fragments) > 0:
        for j, fragment in enumerate(layout.fragments):
          bbox_list.append(_BBox(
            layout_index=i,
            fragment_index=j,
            virtual=False,
            order=0,
            value=fragment.rect.wrapper,
          ))
      else:
        bbox_list.extend(
          self._generate_virtual_lines(
            layout=layout,
            layout_index=i,
            line_height=line_height,
            width=width,
            height=height,
          ),
        )

    if len(bbox_list) > 200:
      # https://github.com/opendatalab/MinerU/blob/980f5c8cd70f22f8c0c9b7b40eaff6f4804e6524/magic_pdf/pdf_parse_union_core_v2.py#L522
      return None

    layoutreader_size = 1000.0
    x_scale = layoutreader_size / float(width)
    y_scale = layoutreader_size / float(height)

    for bbox in bbox_list:
      x0, y0, x1, y1 = self._squeeze(bbox.value, width, height)
      x0 = round(x0 * x_scale)
      y0 = round(y0 * y_scale)
      x1 = round(x1 * x_scale)
      y1 = round(y1 * y_scale)
      bbox.value = (x0, y0, x1, y1)

    bbox_list.sort(key=lambda b: b.value) # 必须排序，乱序传入 layoutreader 会令它无法识别正确顺序
    model = self._get_model()

    with torch.no_grad():
      inputs = boxes2inputs([list(bbox.value) for bbox in bbox_list])
      inputs = prepare_inputs(inputs, model)
     # 将输入数据移动到相同设备
      inputs = {k: v.to(self._device) if isinstance(v, torch.Tensor) else v
                for k, v in inputs.items()}
      logits = model(**inputs).logits.cpu().squeeze(0)
      orders = parse_logits(logits, len(bbox_list))

    sorted_bbox_list = [bbox_list[i] for i in orders]
    for i, bbox in enumerate(sorted_bbox_list):
      bbox.order = i

    return sorted_bbox_list

  def _sort_layouts_and_fragments(self, layouts: list[Layout], bbox_list: list[_BBox]):
    layout_bbox_list: list[list[_BBox]] = [[] for _ in range(len(layouts))]
    for bbox in bbox_list:
      layout_bbox_list[bbox.layout_index].append(bbox)

    layouts_with_median_order: list[tuple[Layout, float]] = []
    for layout_index, bbox_list in enumerate(layout_bbox_list):
      layout = layouts[layout_index]
      orders = [b.order for b in bbox_list] # virtual bbox 保证了 orders 不可能为空
      median_order = self._median(orders)
      layouts_with_median_order.append((layout, median_order))

    for layout, bbox_list in zip(layouts, layout_bbox_list):
      for bbox in bbox_list:
        if not bbox.virtual:
          layout.fragments[bbox.fragment_index].order = bbox.order
      if all(not bbox.virtual for bbox in bbox_list):
        layout.fragments.sort(key=lambda f: f.order)

    layouts_with_median_order.sort(key=lambda x: x[1])
    layouts = [layout for layout, _ in layouts_with_median_order]
    next_fragment_order: int = 0

    for layout in layouts:
      for fragment in layout.fragments:
        fragment.order = next_fragment_order
        next_fragment_order += 1

    return layouts

  def _line_height(self, layouts: list[Layout]) -> float:
    line_height: float = 0.0
    count: int = 0
    for layout in layouts:
      for fragment in layout.fragments:
        _, height = fragment.rect.size
        line_height += height
        count += 1
    if count == 0:
      return 10.0
    return line_height / float(count)

  def _generate_virtual_lines(
        self,
        layout: Layout,
        layout_index: int,
        line_height: float,
        width: int,
        height: int,
      ) -> Generator[_BBox, None, None]:

    # https://github.com/opendatalab/MinerU/blob/980f5c8cd70f22f8c0c9b7b40eaff6f4804e6524/magic_pdf/pdf_parse_union_core_v2.py#L451-L490
    x0, y0, x1, y1 = layout.rect.wrapper
    layout_height = y1 - y0
    layout_weight = x1 - x0
    lines = int(layout_height / line_height)

    if layout_height <= line_height * 2:
      yield _BBox(
        layout_index=layout_index,
        fragment_index=0,
        virtual=True,
        order=0,
        value=(x0, y0, x1, y1),
      )
      return

    elif layout_height <= height * 0.25 or \
         width * 0.5 <= layout_weight or \
         width * 0.25 < layout_weight:
      if layout_weight > width * 0.4:
        lines = 3
      elif layout_weight <= width * 0.25:
        if layout_height / layout_weight > 1.2:  # 细长的不分
          yield _BBox(
            layout_index=layout_index,
            fragment_index=0,
            virtual=True,
            order=0,
            value=(x0, y0, x1, y1),
          )
          return
        else:  # 不细长的还是分成两行
          lines = 2

    lines = max(1, lines)
    line_height = (y1 - y0) / lines
    current_y = y0

    for i in range(lines):
      yield _BBox(
        layout_index=layout_index,
        fragment_index=i,
        virtual=True,
        order=0,
        value=(x0, current_y, x1, current_y + line_height),
      )
      current_y += line_height

  def _median(self, numbers: list[int]) -> float:
    sorted_numbers = sorted(numbers)
    n = len(sorted_numbers)

    # 判断是奇数还是偶数个元素
    if n % 2 == 1:
      # 奇数情况，直接取中间的数
      return float(sorted_numbers[n // 2])
    else:
      # 偶数情况，取中间两个数的平均值
      mid1 = sorted_numbers[n // 2 - 1]
      mid2 = sorted_numbers[n // 2]
      return float((mid1 + mid2) / 2)

  def _squeeze(self, bbox: _BBox, width: int, height: int) -> _BBox:
    x0, y0, x1, y1 = bbox
    x0 = self._squeeze_value(x0, width)
    x1 = self._squeeze_value(x1, width)
    y0 = self._squeeze_value(y0, height)
    y1 = self._squeeze_value(y1, height)
    return x0, y0, x1, y1

  def _squeeze_value(self, position: float, size: int) -> float:
    if position < 0:
      position = 0.0
    if position > size:
      position = float(size)
    return position


from .pix2s import Pix2Struct, Pix2StructTensorRT
from .internvl import InternVL, InternVL_LMDeploy

from transformers import AutoConfig


__ALL_MODELS__ = {
    'Pix2Struct': Pix2Struct,
    'Pix2StructTensorRT': Pix2StructTensorRT,
    'InternVL': InternVL,
    'InternVL_LMDeploy': InternVL_LMDeploy,
}


def get_model_name(model_path):
    model_config = AutoConfig.from_pretrained(
        model_path,
        trust_remote_code=True,
    )

    if 'Pix2Struct' in model_config.architectures[0]:
        model_name = 'Pix2Struct'
    elif 'InternVL' in model_config.architectures[0]:
        model_name = 'InternVL'
    else:
        raise ValueError(f"Unsupported model type: {model_config.architectures[0]}")

    return model_name


def build_model(
        model_ckpt='U4R/StructTable-InternVL2-1B',
        cache_dir=None,
        local_files_only=None,
        **kwargs,
    ):
    model_name = get_model_name(model_ckpt)
    if model_name == 'InternVL' and kwargs.get('lmdeploy', False):
        model_name = 'InternVL_LMDeploy'
    elif model_name == 'Pix2Struct' and kwargs.get('tensorrt_path', None):
        model_name = 'Pix2StructTensorRT'

    model = __ALL_MODELS__[model_name](
        model_ckpt,
        cache_dir=cache_dir,
        local_files_only=local_files_only,
        **kwargs
    )
    return model


import torch

from torch import nn
from transformers import AutoModel, AutoTokenizer, AutoImageProcessor, GenerationConfig

from .conversation import get_conv_template

class InternVL(nn.Module):
    def __init__(
            self,
            model_path='U4R/StructTable-InternVL2-1B',
            max_new_tokens=1024,
            max_time=30,
            flash_attn=True,
            cache_dir=None,
            local_files_only=None,
            **kwargs,
        ):
        super().__init__()
        self.model_path = model_path
        self.max_new_tokens = max_new_tokens
        self.max_generate_time = max_time
        self.flash_attn = flash_attn
        self.cache_dir = cache_dir
        self.local_files_only = local_files_only

        # init model and image processor from ckpt path
        self.init_tokenizer(model_path)
        self.init_image_processor(model_path)
        self.init_model(model_path)

        self.prompt_template = {
            'latex': '<latex>',
            'html': '<html>',
            'markdown': '<markdown>',
        }
        # support output format
        self.supported_output_format = ['latex', 'html', 'markdown']

    def init_model(self, model_path):
        self.model = AutoModel.from_pretrained(
            pretrained_model_name_or_path=model_path,
            trust_remote_code=True,
            torch_dtype=torch.bfloat16,
            low_cpu_mem_usage=True,
            use_flash_attn=self.flash_attn,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )
        self.model.eval()

    def init_image_processor(self, image_processor_path):
        self.image_processor = AutoImageProcessor.from_pretrained(
            pretrained_model_name_or_path=image_processor_path,
            trust_remote_code=True,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )

    def init_tokenizer(self, tokenizer_path):
        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=tokenizer_path,
            trust_remote_code=True,
            use_fast=False,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )

        self.image_context_token = '<IMG_CONTEXT>'
        self.image_token_num = 256
        self.image_start_token = '<img>'
        self.image_end_token = '</img>'
        self.img_context_token_id = self.tokenizer.convert_tokens_to_ids(self.image_context_token)

    def format_image_tokens(self, path_num):
        return f'{self.image_start_token}{self.image_context_token* self.image_token_num * path_num}{self.image_end_token}'

    def forward(self, images, output_format='latex', **kwargs):
        # process image to tokens
        if not isinstance(images, list):
            images = [images]

        pixel_values_list = []
        for image in images:
            path_images = self.dynamic_preprocess(
                image, image_size=448, max_num=12
            )
            pixel_values = self.image_processor(
                path_images,
                return_tensors='pt'
            )['pixel_values'].to(torch.bfloat16)
            pixel_values_list.append(pixel_values)

        batch_size = len(pixel_values_list)
        conversation_list = []
        for bs_idx in range(batch_size):
            pixel_values= pixel_values_list[bs_idx].to(torch.bfloat16)

            image_tokens = self.format_image_tokens(pixel_values.shape[0])
            question = '<image>\n' + self.prompt_template[output_format]
            answer = None

            template = get_conv_template(self.model.config.template)
            template.append_message(template.roles[0], question)
            template.append_message(template.roles[1], answer)
            conversation = template.get_prompt()
            conversation = conversation.replace('<image>', image_tokens, 1)
            conversation_list.append(conversation)

        device = next(self.parameters()).device
        self.tokenizer.padding_side = 'left'
        model_inputs = self.tokenizer(
            conversation_list,
            return_tensors='pt',
            padding=True,
            max_length=self.tokenizer.model_max_length,
            truncation=True,
        ).to(device)
        pixel_values = torch.cat(pixel_values_list, axis=0).to(device)

        # generation config
        generation_config = dict(
            max_new_tokens=self.max_new_tokens,
            max_time=self.max_generate_time,
            img_context_token_id=self.img_context_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
            eos_token_id=self.tokenizer.eos_token_id,
            do_sample=False,
            no_repeat_ngram_size=20,
        )

        # generate text from image tokens
        model_output = self.model.generate(
            pixel_values=pixel_values,
            input_ids=model_inputs.input_ids,
            attention_mask=model_inputs.attention_mask,
            **generation_config,
            # **kwargs
        )

        batch_decode_texts = self.tokenizer.batch_decode(
            model_output,
            skip_special_tokens=True
        )
        return batch_decode_texts

    def find_closest_aspect_ratio(self, aspect_ratio, target_ratios, width, height, image_size):
        best_ratio_diff = float('inf')
        best_ratio = (1, 1)
        area = width * height
        for ratio in target_ratios:
            target_aspect_ratio = ratio[0] / ratio[1]
            ratio_diff = abs(aspect_ratio - target_aspect_ratio)
            if ratio_diff < best_ratio_diff:
                best_ratio_diff = ratio_diff
                best_ratio = ratio
            elif ratio_diff == best_ratio_diff:
                if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                    best_ratio = ratio
        return best_ratio

    def dynamic_preprocess(self, image, min_num=1, max_num=12, image_size=448, use_thumbnail=True):
        orig_width, orig_height = image.size
        aspect_ratio = orig_width / orig_height

        # calculate the existing image aspect ratio
        target_ratios = set(
            (i, j) for n in range(min_num, max_num + 1) for i in range(1, n + 1) for j in range(1, n + 1) if
            i * j <= max_num and i * j >= min_num)
        target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

        # find the closest aspect ratio to the target
        target_aspect_ratio = self.find_closest_aspect_ratio(
            aspect_ratio, target_ratios, orig_width, orig_height, image_size)

        # calculate the target width and height
        target_width = image_size * target_aspect_ratio[0]
        target_height = image_size * target_aspect_ratio[1]
        blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

        # resize the image
        resized_img = image.resize((target_width, target_height))
        processed_images = []
        for i in range(blocks):
            box = (
                (i % (target_width // image_size)) * image_size,
                (i // (target_width // image_size)) * image_size,
                ((i % (target_width // image_size)) + 1) * image_size,
                ((i // (target_width // image_size)) + 1) * image_size
            )
            # split the image
            split_img = resized_img.crop(box)
            processed_images.append(split_img)
        assert len(processed_images) == blocks
        if use_thumbnail and len(processed_images) != 1:
            thumbnail_img = image.resize((image_size, image_size))
            processed_images.append(thumbnail_img)
        return processed_images



import os
import time
import json

import torch
import torch.nn as nn

try:
    import tensorrt_llm
    import tensorrt as trt
    import tensorrt_llm.profiler as profiler

    from tensorrt_llm._utils import str_dtype_to_trt, torch_to_numpy
    from tensorrt_llm.lora_manager import LoraManager
    from tensorrt_llm.runtime import Session, TensorInfo, ModelConfig, SamplingConfig
except:
    print("\033[93mimport tensorrt_llm failed, if do not use tensorrt, ignore this message\033[0m")

from typing import List
from transformers import AutoProcessor, AutoTokenizer, AutoConfig


def trt_dtype_to_torch(dtype):
    if dtype == trt.float16:
        return torch.float16
    elif dtype == trt.float32:
        return torch.float32
    elif dtype == trt.int32:
        return torch.int32
    elif dtype == trt.bfloat16:
        return torch.bfloat16
    else:
        raise TypeError("%s is not supported" % dtype)


class Pix2StructTensorRT(nn.Module):

    def __init__(
            self,
            model_path,
            tensorrt_path,
            batch_size=1,
            max_new_tokens=4096,
            cache_dir=None,
            local_files_only=None,
            **kwargs,
        ):
        self.model_ckpt_path = model_path
        self.tensorrt_path = tensorrt_path
        self.batch_size = batch_size
        self.max_new_tokens = max_new_tokens
        self.cache_dir = cache_dir
        self.local_files_only = local_files_only

        self.llm_engine_path = os.path.join(tensorrt_path, 'llm_engines')
        self.visual_engine_path = os.path.join(tensorrt_path, 'visual_engines')

        device_id = torch.cuda.current_device() % torch.cuda.device_count()
        self.device_id = device_id
        self.device = "cuda:%d" % (device_id)

        self.stream = torch.cuda.Stream(torch.cuda.current_device())
        torch.cuda.set_stream(self.stream)

        # parse model type from visual engine config
        with open(os.path.join(self.visual_engine_path, "config.json"),
                  "r") as f:
            config = json.load(f)
        self.model_type = config['builder_config']['model_type']
        self.vision_precision = config['builder_config']['precision']

        self.vision_precision = 'float16'
        self.decoder_llm = not (
            't5' in self.model_type
            or self.model_type in ['nougat', 'pix2struct', 'StructEqTable']
        )  # BLIP2-T5, pix2struct and Nougat are using encoder-decoder models as LLMs

        self.profiling_iterations = 20

        self.init_image_encoder()
        self.init_tokenizer()
        self.init_llm()
        self.init_image_processor()

        self.special_str_list = ['\\midrule', '\\hline']
        self.supported_output_format = ['latex']

    def postprocess_latex_code(self, code):
        for special_str in self.special_str_list:
            code = code.replace(special_str, special_str + ' ')
        return code

    def init_image_processor(self):
        self.data_processor = AutoProcessor.from_pretrained(
            pretrained_model_name_or_path=self.model_ckpt_path,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )

    def init_tokenizer(self):
        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=self.model_ckpt_path,
            use_fast=True,
            use_legacy=False,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )
        # self.tokenizer.padding_side = "right"

    def init_image_encoder(self):
        vision_encoder_path = os.path.join(self.visual_engine_path,
                                           'visual_encoder.engine')
        with open(vision_encoder_path, 'rb') as f:
            engine_buffer = f.read()
        self.visual_encoder_session = Session.from_serialized_engine(
            engine_buffer)

    def init_llm(self):

        self.model = TRTLLMEncDecModel.from_engine(
            os.path.basename(self.model_ckpt_path),
            self.llm_engine_path,
            skip_encoder=self.model_type in ['nougat', 'pix2struct', 'StructEqTable'],
            debug_mode=False,
            stream=self.stream)

        self.model_config = self.model.decoder_model_config
        self.runtime_mapping = self.model.decoder_runtime_mapping

    def __call__(self, image, **kwargs):
        # process image to tokens
        image_tokens = self.data_processor.image_processor(
            images=image,
            return_tensors='pt',
        )

        for k, v in image_tokens.items():
            image_tokens[k] = v.cuda()

        model_output = self.run(
            flattened_patches=image_tokens['flattened_patches'],
            attention_mask=image_tokens['attention_mask'],
            max_new_tokens=self.max_new_tokens
        )

        # postprocess
        latex_codes = []
        for i, code in enumerate(model_output):
            latex_codes.append(self.postprocess_latex_code(code[0]))

        return latex_codes

    def preprocess(self, warmup, pre_prompt, post_prompt, image,
                   attention_mask):
        if not warmup:
            profiler.start("Vision")

        visual_features, visual_atts = self.get_visual_features(
            torch.stack(image['image_patches'], dim=0)
            if self.model_type == 'fuyu' else image, attention_mask)

        if not warmup:
            profiler.stop("Vision")

        pre_input_ids = self.tokenizer(pre_prompt,
                                        return_tensors="pt",
                                        padding=True).input_ids
        if post_prompt[0] is not None:
            post_input_ids = self.tokenizer(post_prompt,
                                            return_tensors="pt",
                                            padding=True).input_ids
            length = pre_input_ids.shape[1] + post_input_ids.shape[
                1] + visual_atts.shape[1]
        else:
            post_input_ids = None
            length = pre_input_ids.shape[1] + visual_atts.shape[1]

        input_lengths = torch.IntTensor([length] * 1).to(
            torch.int32)

        input_ids, ptuning_args = self.setup_fake_prompts(
            visual_features, pre_input_ids, post_input_ids, input_lengths)

        return input_ids, input_lengths, ptuning_args, visual_features

    def generate(self, pre_prompt, post_prompt, image, decoder_input_ids,
                 max_new_tokens, attention_mask, warmup):
        if not warmup:
            profiler.start("Generate")

        input_ids, input_lengths, ptuning_args, visual_features = self.preprocess(
            warmup, pre_prompt, post_prompt, image, attention_mask)

        if warmup: return None

        profiler.start("LLM")

        # Trim encoder input_ids to match visual features shape
        ids_shape = (self.batch_size, visual_features.shape[1])

        input_ids = torch.ones(ids_shape, dtype=torch.int32)

        output_ids = self.model.generate(
            input_ids,
            decoder_input_ids,
            max_new_tokens,
            num_beams=1,
            bos_token_id=self.tokenizer.bos_token_id,
            pad_token_id=self.tokenizer.pad_token_id,
            eos_token_id=self.tokenizer.eos_token_id,
            debug_mode=False,
            prompt_embedding_table=ptuning_args[0],
            prompt_tasks=ptuning_args[1],
            prompt_vocab_size=ptuning_args[2],
            attention_mask=attention_mask)

        # Reset input_lengths to match decoder_input_ids
        input_lengths = torch.ones(input_lengths.shape,
                                    dtype=input_lengths.dtype)
        profiler.stop("LLM")

        if tensorrt_llm.mpi_rank() == 0:
            # Extract a list of tensors of shape beam_width x output_ids.
            output_beams_list = [
                self.tokenizer.batch_decode(
                    output_ids[batch_idx, :, input_lengths[batch_idx]:],
                    skip_special_tokens=True)
                for batch_idx in range(self.batch_size)
            ]

            stripped_text = [[
                output_beams_list[batch_idx][beam_idx].strip()
                for beam_idx in range(1)
            ] for batch_idx in range(self.batch_size)]
            profiler.stop("Generate")
            return stripped_text
        else:
            profiler.stop("Generate")
            return None

    def get_visual_features(self, image, attention_mask):
        visual_features = {
            'input':
            image.to(
                tensorrt_llm._utils.str_dtype_to_torch(self.vision_precision))
        }
        if attention_mask is not None:
            visual_features['attention_mask'] = attention_mask
        tensor_info = [
            TensorInfo('input', str_dtype_to_trt(self.vision_precision),
                       image.shape)
        ]
        if attention_mask is not None:
            tensor_info.append(
                TensorInfo('attention_mask', trt.DataType.INT32,
                           attention_mask.shape))
        visual_output_info = self.visual_encoder_session.infer_shapes(
            tensor_info)
        visual_outputs = {
            t.name: torch.empty(tuple(t.shape),
                                dtype=trt_dtype_to_torch(t.dtype),
                                device=image.device)
            for t in visual_output_info
        }

        ok = self.visual_encoder_session.run(visual_features, visual_outputs,
                                             self.stream.cuda_stream)
        assert ok, "Runtime execution failed for vision encoder session"
        self.stream.synchronize()

        image_embeds = visual_outputs['output']
        image_atts = torch.ones(image_embeds.size()[:-1],
                                dtype=torch.long).to(image.device)

        return image_embeds, image_atts

    def setup_fake_prompts(self, visual_features, pre_input_ids, post_input_ids,
                           input_lengths):
        # Assemble fake prompts which points to image embedding actually
        fake_prompt_id = torch.arange(
            self.model_config.vocab_size, self.model_config.vocab_size +
            visual_features.shape[0] * visual_features.shape[1])
        fake_prompt_id = fake_prompt_id.reshape(visual_features.shape[0],
                                                visual_features.shape[1])

        if post_input_ids is not None:
            input_ids = [pre_input_ids, fake_prompt_id, post_input_ids]
        else:
            input_ids = [fake_prompt_id, pre_input_ids]

        input_ids = torch.cat(input_ids, dim=1).contiguous().to(torch.int32)

        if self.decoder_llm or self.runtime_mapping.is_first_pp_rank():
            ptuning_args = self.ptuning_setup(visual_features, input_ids,
                                              input_lengths)
        else:
            ptuning_args = [None, None, None]

        return input_ids, ptuning_args

    def ptuning_setup(self, prompt_table, input_ids, input_lengths):
        hidden_size = self.model_config.hidden_size * self.runtime_mapping.tp_size
        if prompt_table is not None:
            task_vocab_size = torch.tensor(
                [prompt_table.shape[1]],
                dtype=torch.int32,
            ).cuda()
            prompt_table = prompt_table.view(
                (prompt_table.shape[0] * prompt_table.shape[1],
                 prompt_table.shape[2]))
            assert prompt_table.shape[
                1] == hidden_size, "Prompt table dimensions do not match hidden size"

            prompt_table = prompt_table.cuda().to(
                dtype=tensorrt_llm._utils.str_dtype_to_torch(
                    self.model_config.dtype))
        else:
            prompt_table = torch.empty([1, hidden_size]).cuda()
            task_vocab_size = torch.zeros([1]).cuda()

        if self.model_config.remove_input_padding:
            tasks = torch.zeros([torch.sum(input_lengths)],
                                dtype=torch.int32).cuda()
            if self.decoder_llm: tasks = tasks.unsqueeze(0)
        else:
            tasks = torch.zeros(input_ids.shape, dtype=torch.int32).cuda()

        return [prompt_table, tasks, task_vocab_size]

    def setup_inputs(self, input_text, raw_image):
        attention_mask = None

        image_processor = AutoProcessor.from_pretrained(
            pretrained_model_name_or_path=self.model_ckpt_path,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )
        if input_text is None:
            input_text = ""
        inputs = image_processor(
            images=raw_image,
            text=input_text,
            return_tensors="pt",
        )
        image = inputs['flattened_patches']
        image = image.expand(self.batch_size, -1, -1).contiguous()
        attention_mask = inputs['attention_mask'].to(self.device).to(
            torch.int)
        attention_mask = attention_mask.expand(self.batch_size,
                                                -1).contiguous()
        pre_prompt = ""
        post_prompt = None

        # Repeat inputs to match batch size
        pre_prompt = [pre_prompt] * self.batch_size
        post_prompt = [post_prompt] * self.batch_size
        image = image.to(self.device)

        # Generate decoder_input_ids for enc-dec models
        # Custom prompts can be added as:
        # decoder_input_ids = model.tokenizer(decoder_prompt).input_ids
        if self.decoder_llm:
            decoder_input_ids = None
        else:
            config = AutoConfig.from_pretrained(
                pretrained_model_name_or_path=self.model_ckpt_path,
                cache_dir=self.cache_dir,
                local_files_only=self.local_files_only,
            )
            decoder_start_id = config.decoder_start_token_id  # T5
            if decoder_start_id is None:
                decoder_start_id = config.decoder.bos_token_id  # Nougat

            decoder_input_ids = torch.IntTensor([[decoder_start_id]])
            decoder_input_ids = decoder_input_ids.repeat((self.batch_size, 1))

        return input_text, pre_prompt, post_prompt, image, decoder_input_ids, attention_mask

    def run(self, flattened_patches, attention_mask, max_new_tokens):
        # input_text, pre_prompt, post_prompt, processed_image, decoder_input_ids, attention_mask = self.setup_inputs(
        #     None, raw_image)
        pre_prompt = [""] * self.batch_size
        post_prompt = [None] * self.batch_size
        config = AutoConfig.from_pretrained(
            pretrained_model_name_or_path=self.model_ckpt_path,
            cache_dir=self.cache_dir,
            local_files_only=self.local_files_only,
        )
        decoder_start_id = config.decoder_start_token_id  # T5
        decoder_input_ids = torch.IntTensor([[decoder_start_id]])
        decoder_input_ids = decoder_input_ids.repeat((self.batch_size, 1))

        processed_image = flattened_patches.expand(self.batch_size, -1, -1).contiguous()
        attention_mask = attention_mask.to(self.device).to(torch.int)
        attention_mask = attention_mask.expand(self.batch_size,-1).contiguous()

        self.generate(pre_prompt,
                       post_prompt,
                       processed_image,
                       decoder_input_ids,
                       max_new_tokens,
                       attention_mask=attention_mask,
                       warmup=True)
        # num_iters = self.profiling_iterations if self.args.run_profiling else 1
        num_iters = 1
        # print(num_iters)
        for _ in range(num_iters):
            output_text = self.generate(pre_prompt,
                                         post_prompt,
                                         processed_image,
                                         decoder_input_ids,
                                         max_new_tokens,
                                         attention_mask=attention_mask,
                                         warmup=False)
        # if self.runtime_rank == 0:
        #     self.print_result(input_text, output_text)
        return output_text


def read_config(config_path):
    with open(config_path, "r") as f:
        config = json.load(f)

    builder_config = config['build_config']
    plugin_config = builder_config['plugin_config']
    pretrained_config = config['pretrained_config']
    lora_config = builder_config['lora_config']
    auto_parallel_config = builder_config['auto_parallel_config']
    use_gpt_attention_plugin = plugin_config["gpt_attention_plugin"]
    remove_input_padding = plugin_config["remove_input_padding"]
    use_lora_plugin = plugin_config["lora_plugin"]
    tp_size = pretrained_config['mapping']['tp_size']
    pp_size = pretrained_config['mapping']['pp_size']
    gpus_per_node = auto_parallel_config['gpus_per_node']
    world_size = tp_size * pp_size
    assert world_size == tensorrt_llm.mpi_world_size(), \
        f'Engine world size ({world_size}) != Runtime world size ({tensorrt_llm.mpi_world_size()})'
    num_heads = pretrained_config["num_attention_heads"]
    hidden_size = pretrained_config["hidden_size"]
    head_size = pretrained_config["head_size"]
    vocab_size = pretrained_config["vocab_size"]
    max_batch_size = builder_config["max_batch_size"]
    max_beam_width = builder_config["max_beam_width"]
    num_layers = pretrained_config["num_hidden_layers"]
    num_kv_heads = pretrained_config.get('num_kv_heads', num_heads)

    assert (num_heads % tp_size) == 0
    num_heads = num_heads // tp_size
    hidden_size = hidden_size // tp_size
    num_kv_heads = (num_kv_heads + tp_size - 1) // tp_size

    cross_attention = pretrained_config["architecture"] == "DecoderModel"
    skip_cross_qkv = pretrained_config.get('skip_cross_qkv', False)
    has_position_embedding = pretrained_config["has_position_embedding"]
    has_token_type_embedding = hasattr(pretrained_config, "type_vocab_size")
    use_custom_all_reduce = plugin_config.get('use_custom_all_reduce', False)
    dtype = pretrained_config["dtype"]

    paged_kv_cache = plugin_config['paged_kv_cache']
    tokens_per_block = plugin_config['tokens_per_block']

    gather_context_logits = builder_config.get('gather_context_logits', False)
    gather_generation_logits = builder_config.get('gather_generation_logits',
                                                  False)
    max_prompt_embedding_table_size = builder_config.get(
        'max_prompt_embedding_table_size', 0)

    model_config = ModelConfig(
        num_heads=num_heads,
        num_kv_heads=num_kv_heads,
        hidden_size=hidden_size,
        head_size=head_size,
        max_batch_size=max_batch_size,
        max_beam_width=max_beam_width,
        vocab_size=vocab_size,
        num_layers=num_layers,
        gpt_attention_plugin=use_gpt_attention_plugin,
        remove_input_padding=remove_input_padding,
        paged_kv_cache=paged_kv_cache,
        tokens_per_block=tokens_per_block,
        cross_attention=cross_attention,
        has_position_embedding=has_position_embedding,
        has_token_type_embedding=has_token_type_embedding,
        use_custom_all_reduce=use_custom_all_reduce,
        dtype=dtype,
        gather_context_logits=gather_context_logits,
        gather_generation_logits=gather_generation_logits,
        max_prompt_embedding_table_size=max_prompt_embedding_table_size,
        lora_plugin=use_lora_plugin,
        lora_target_modules=lora_config.get('lora_target_modules'),
        trtllm_modules_to_hf_modules=lora_config.get(
            'trtllm_modules_to_hf_modules'),
        skip_cross_qkv=skip_cross_qkv,
    )

    return model_config, tp_size, pp_size, gpus_per_node, dtype


class Mapping(object):
    def __init__(
            self,
            world_size=1,
            rank=0,
            gpus_per_node=8,
            tp_size=1,
            pp_size=1,
            moe_tp_size=-1,  # -1 means no moe
            moe_ep_size=-1):  # -1 means no moe
        # set default values for non-moe cases
        if moe_tp_size == -1:
            moe_tp_size = tp_size
            moe_ep_size = 1

        if pp_size * tp_size != world_size:
            raise ValueError(
                f"world_size must equal to pp_size * tp_size, but got {world_size} != {pp_size} * {tp_size}"
            )

        moe_tp_ep_size = moe_tp_size * moe_ep_size
        if moe_tp_ep_size != tp_size:
            raise ValueError(
                f"tp_size must equal to moe_tp_size * moe_ep_size, but got {tp_size} != {moe_tp_size} * {moe_ep_size}"
            )

        self.tp_size = tp_size
        self.pp_size = pp_size
        self.moe_tp_size = moe_tp_size
        self.moe_ep_size = moe_ep_size
        self.world_size = world_size
        self.rank = rank
        self.gpus_per_node = gpus_per_node

        self.pp_groups = []
        self.tp_groups = []
        self.moe_tp_groups = []
        self.moe_ep_groups = []

        # init pp group
        for i in range(tp_size):
            ranks = range(i+ self.rank, world_size+ self.rank, tp_size)
            self.pp_groups.append(list(ranks))

        # init tp group
        for i in range(pp_size):
            ranks = range(i * tp_size + self.rank, (i + 1) * tp_size + self.rank)
            self.tp_groups.append(list(ranks))

        # init moe tp group
        for i in range(pp_size):
            for j in range(moe_ep_size):
                ranks = range(i * moe_tp_ep_size + j, (i + 1) * moe_tp_ep_size,
                              moe_ep_size)
                self.moe_tp_groups.append(list(ranks))

        # init moe ep group
        for i in range(pp_size):
            for j in range(moe_tp_size):
                ranks = range(i * moe_tp_ep_size + j * moe_ep_size,
                              i * moe_tp_ep_size + (j + 1) * moe_ep_size)
                self.moe_ep_groups.append(list(ranks))

        # self.pp_rank = self.rank // self.tp_size
        # self.tp_rank = self.rank % self.tp_size
        self.pp_rank = 0
        self.tp_rank = 0
        self.moe_tp_rank = self.tp_rank // self.moe_ep_size
        self.moe_ep_rank = self.tp_rank % self.moe_ep_size

        # self.tp_group = self.tp_groups[self.pp_rank]
        # self.pp_group = self.pp_groups[self.tp_rank]
        self.moe_tp_group = self.moe_tp_groups[self.pp_rank * moe_ep_size +
                                               self.moe_ep_rank]
        self.moe_ep_group = self.moe_ep_groups[self.pp_rank * moe_tp_size +
                                               self.moe_tp_rank]

        self.node_rank = self.rank // self.gpus_per_node
        self.local_rank = self.rank % self.gpus_per_node

    def get_node_rank(self, rank: int):
        return rank // self.gpus_per_node

    def get_local_rank(self, rank: int):
        return rank % self.gpus_per_node

    def has_tp(self):
        return self.tp_size > 1

    def is_last_pp_rank(self):
        return self.pp_rank == self.pp_size - 1

    def is_first_pp_rank(self):
        return self.pp_rank == 0

    def has_pp(self):
        return self.pp_size > 1

    def prev_pp_rank(self):
        p = self.rank - self.tp_size
        if p < 0:
            p = p + self.world_size
        return p

    def next_pp_rank(self):
        p = self.rank + self.tp_size
        if p >= self.world_size:
            p = p - self.world_size
        return p

    def has_moe_tp(self):
        return self.moe_tp_size > 1

    def has_moe_ep(self):
        return self.moe_ep_size > 1

    def pp_layers(self, num_layers: int) -> List[int]:
        layers_per_pipeline_stage = num_layers // self.pp_size
        layers_range = range(self.pp_rank * layers_per_pipeline_stage,
                             (self.pp_rank + 1) * layers_per_pipeline_stage)
        return list(layers_range)

    def ep_experts(self, num_experts: int) -> List[int]:
        experts_per_rank = num_experts // self.moe_ep_size
        experts_range = range(self.moe_ep_rank * experts_per_rank,
                              (self.moe_ep_rank + 1) * experts_per_rank)
        return list(experts_range)


def get_engine_name(rank):
    return 'rank{}.engine'.format(rank)

class TRTLLMEncDecModel:

    def __init__(
        self,
        engine_name,
        engine_dir,
        lora_dir=None,
        lora_task_uids=None,
        debug_mode=False,
        skip_encoder=False,
        stream: torch.cuda.Stream = None,
    ):
        # in multi-node setup, it's important to set_device at the very beginning so .to('cuda') refers to current device
        # accordingly, all input & output tensors should be moved to current device
        # otherwise, it's default to 'cuda:0'

        # self.runtime_rank = tensorrt_llm.mpi_rank()
        self.device_id = torch.cuda.current_device()
        # torch.cuda.set_device(device_id)
        self.device = torch.cuda.current_device()
        self.skip_encoder = skip_encoder
        self.lora_task_uids = lora_task_uids

        # when enc-dec runs by itself, stream can be None and we create new stream here
        # when enc-dec has to run as a component in a bigger workflow (e.g., multimodal), earlier components in the workflow may have results in its stream, which we should pass that stream in to avoid unnecessary stream sync
        self.stream = stream
        if self.stream is None:
            self.stream = torch.cuda.Stream(self.device)
        torch.cuda.set_stream(self.stream)

        def engine_setup(component):
            # model config
            config_path = os.path.join(engine_dir, component, "config.json")
            model_config, tp_size, pp_size, gpus_per_node, dtype = read_config(
                config_path)

            # MGMN config
            world_size = tp_size * pp_size
            # runtime_rank = tensorrt_llm.mpi_rank()
            runtime_rank = torch.cuda.current_device()
            # assert runtime_rank < world_size, "Runtime GPU rank exceeds MPI world size. Did you launch more MPI processes than required?"
            # runtime_mapping = tensorrt_llm.Mapping(world_size,
            #                                        runtime_rank,
            #                                        tp_size=tp_size,
            #                                        pp_size=pp_size,
            #                                        gpus_per_node=gpus_per_node)
            # tensorrt_llm.Mapping
            runtime_mapping = Mapping(world_size,
                                      runtime_rank,
                                      tp_size=tp_size,
                                      pp_size=pp_size,
                                      gpus_per_node=gpus_per_node)
            # load engine
            # engine_fname = get_engine_name(runtime_rank)
            engine_fname = get_engine_name(0)
            with open(os.path.join(engine_dir, component, engine_fname), "rb") as f:
                engine_buffer = f.read()

            return model_config, runtime_mapping, engine_buffer

        # Note: encoder and decoder doesn't necessarily have the same TP & PP config

        if not skip_encoder:
            self.encoder_model_config, self.encoder_runtime_mapping, encoder_engine_buffer = engine_setup(
                component='encoder')

            self.nccl_comm = None
            if self.encoder_runtime_mapping.has_pp():
                # for Pipeline Parallelism in encoder
                self.nccl_comm = torch.classes.trtllm.NcclCommunicatorOp(
                    self.encoder_runtime_mapping.tp_size,
                    self.encoder_runtime_mapping.pp_size,
                    self.encoder_runtime_mapping.rank)

            # session setup
            self.encoder_session = tensorrt_llm.runtime.Session.from_serialized_engine(
                encoder_engine_buffer)

            # encoder lora manager setup
            if self.encoder_model_config.lora_plugin:
                self.encoder_lora_manager = LoraManager()
                # TODO: this is only for bart
                self.encoder_lora_manager.load_from_hf(
                    model_dirs=lora_dir,
                    model_config=self.encoder_model_config,
                    runtime_mapping=self.encoder_runtime_mapping,
                    component='encoder',
                )
            else:
                self.encoder_lora_manager = None
        else:
            self.encoder_model_config, self.encoder_runtime_mapping, encoder_engine_buffer = None, None, None
            self.nccl_comm, self.encoder_session = None, None

        self.decoder_model_config, self.decoder_runtime_mapping, decoder_engine_buffer = engine_setup(
            component='decoder')

        self.decoder_session = tensorrt_llm.runtime.GenerationSession(
            self.decoder_model_config,
            decoder_engine_buffer,
            self.decoder_runtime_mapping,
            debug_mode=debug_mode)

        # decoder lora manager setup
        if self.decoder_model_config.lora_plugin:
            self.decoder_lora_manager = LoraManager()
            # TODO: this is only for bart
            self.decoder_lora_manager.load_from_hf(
                model_dirs=lora_dir,
                model_config=self.decoder_model_config,
                runtime_mapping=self.decoder_runtime_mapping,
                component='decoder',
            )
        else:
            self.decoder_lora_manager = None

    @classmethod
    def from_engine(cls,
                    engine_name,
                    engine_dir,
                    lora_dir=None,
                    lora_task_uids=None,
                    debug_mode=False,
                    skip_encoder=False,
                    stream=None):
        return cls(engine_name,
                   engine_dir,
                   lora_dir,
                   lora_task_uids,
                   debug_mode=debug_mode,
                   skip_encoder=skip_encoder,
                   stream=stream)

    def process_input(self,
                      input_ids,
                      remove_input_padding=False,
                      pad_token_id=0,
                      prompt_tasks=None):
        if remove_input_padding:
            # in remove padding mode --> flatten input, calculate actual length and max length
            # Note: 1st token should never be removed, even if it is pad_token_id
            first_ids = input_ids[:, 0]
            input_ids = input_ids[:, 1:]
            input_lengths = 1 + (input_ids != pad_token_id).sum(dim=1).type(
                torch.IntTensor).to(self.device)  # [batch_size]
            new_ids = []
            for i in range(len(input_ids)):
                row = input_ids[i, :]
                row = row[row != pad_token_id]
                new_ids.append(
                    torch.cat(
                        (torch.IntTensor([first_ids[i]]).to(self.device), row)))
            input_ids = torch.cat(new_ids)  # [num_tokens]
            if prompt_tasks is not None:
                prompt_tasks = prompt_tasks[:input_ids.shape[0]]
        else:
            # in padding mode --> keep input, just calculate actual length and max length
            # Note: 1st token should always count, even if it is pad_token_id. e.g., decoder start id in enc-dec models could be a single pad_token_id, we should count
            input_lengths = torch.tensor(
                1 + (input_ids[:, 1:] != pad_token_id).sum(dim=1).type(
                    torch.IntTensor).to(self.device),
                dtype=torch.int32,
                device=self.device)
        max_input_length = torch.max(input_lengths).item()
        return input_ids, input_lengths, max_input_length, prompt_tasks

    def encoder_run(self,
                    input_ids,
                    input_lengths,
                    max_input_length,
                    position_ids=None,
                    token_type_ids=None,
                    debug_mode=False,
                    prompt_embedding_table=None,
                    prompt_tasks=None,
                    prompt_vocab_size=None,
                    attention_mask=None):

        # each engine has hidden_dim/TP, don't forget to multiply TP
        hidden_size = self.encoder_model_config.hidden_size * self.encoder_runtime_mapping.tp_size
        if input_ids.dim() == 1:
            hidden_states_shape = (input_ids.shape[0], hidden_size
                                   )  # [num_tokens,D]
        else:
            hidden_states_shape = (input_ids.shape[0], input_ids.shape[1],
                                   hidden_size)  # [BS,seqlen,D]
        hidden_states_dtype = lambda name: trt_dtype_to_torch(
            self.encoder_session.engine.get_tensor_dtype(name))

        # input tensors. only first PP rank has id input, others are hidden_states input
        inputs = {}
        if self.encoder_runtime_mapping.is_first_pp_rank():
            inputs['input_ids'] = input_ids.contiguous()
            if self.encoder_model_config.has_position_embedding:
                if position_ids is None:
                    if self.encoder_model_config.remove_input_padding:
                        position_ids = [
                            torch.arange(sample_length,
                                         dtype=torch.int32,
                                         device=input_ids.device)
                            for sample_length in torch_to_numpy(input_lengths)
                        ]
                        position_ids = torch.cat(position_ids)
                    else:
                        bsz, seq_len = input_ids.shape[:2]
                        position_ids = torch.arange(
                            seq_len, dtype=torch.int32,
                            device=input_ids.device).expand(bsz, -1)
                inputs['position_ids'] = position_ids.contiguous()
            if self.encoder_model_config.has_token_type_embedding:
                inputs['token_type_ids'] = token_type_ids.contiguous()

            if self.encoder_model_config.max_prompt_embedding_table_size > 0:
                inputs[
                    'prompt_embedding_table'] = prompt_embedding_table.contiguous(
                    )
                inputs['tasks'] = prompt_tasks.contiguous()
                inputs['prompt_vocab_size'] = prompt_vocab_size.contiguous()
        else:
            # just need a placeholder, engine will call NCCL to recv and fill data from previous rank
            inputs['hidden_states_input'] = torch.empty(
                hidden_states_shape,
                dtype=hidden_states_dtype('hidden_states_input'),
                device=self.device).contiguous()
        if attention_mask is not None and not self.encoder_model_config.gpt_attention_plugin:
            inputs['attention_mask'] = attention_mask.contiguous()

        inputs['input_lengths'] = input_lengths
        # use shape info to pass max length info in remove padding mode
        inputs['max_input_length'] = torch.empty(
            (max_input_length, ),
            dtype=hidden_states_dtype('max_input_length'),
            device=self.device).contiguous()
        batch_size = input_lengths.size(0)
        inputs['host_request_types'] = torch.IntTensor([0] *
                                                       batch_size).to('cpu')
        if self.encoder_model_config.remove_input_padding:
            inputs['host_context_lengths'] = input_lengths.to('cpu')

        if self.encoder_model_config.lora_plugin and self.encoder_lora_manager is not None:
            inputs.update(
                self.encoder_lora_manager.input_buffers(
                    self.lora_task_uids,
                    self.encoder_runtime_mapping,
                    self.encoder_model_config.num_layers,
                ))

        # Note: runtime.Session's run() method will set input/output tensor address, here we only need to provide tensor shape
        self.encoder_session.set_shapes(inputs)

        # output tensors. only last PP rank final encoder output, others are intermediate hidden_states output. Need broadcast later
        outputs = {}
        if self.encoder_runtime_mapping.is_last_pp_rank():
            outputs['encoder_output'] = torch.empty(
                hidden_states_shape,
                dtype=hidden_states_dtype('encoder_output'),
                device=self.device).contiguous()
        else:
            outputs['hidden_states_output'] = torch.empty(
                hidden_states_shape,
                dtype=hidden_states_dtype('hidden_states_output'),
                device=self.device).contiguous()

        # -------------------------------------------
        if debug_mode:
            engine = self.encoder_session.engine
            context = self.encoder_session.context
            # setup debugging buffer for the encoder
            for i in range(self.encoder_session.engine.num_io_tensors):
                name = engine.get_tensor_name(i)
                if engine.get_tensor_mode(
                        name
                ) == trt.TensorIOMode.OUTPUT and name not in outputs.keys():
                    dtype = engine.get_tensor_dtype(name)
                    shape = context.get_tensor_shape(name)
                    outputs[name] = torch.zeros(tuple(shape),
                                                dtype=trt_dtype_to_torch(dtype),
                                                device=self.device)
                    context.set_tensor_address(name, outputs[name].data_ptr())
        # -------------------------------------------

        # TRT session run
        # Note: need cuda stream ID, not a torch Stream
        ok = self.encoder_session.run(inputs, outputs, self.stream.cuda_stream)
        assert ok, "Runtime execution failed"
        self.stream.synchronize()

        # Tensor Parallelism is handled by model/engine definition
        # But we need to broadcast among PP group at the end of encoder's Pipeline Parallelism
        # After this, all ranks should recv the encoder output, and world might be re-configured using decoder's TP-PP config
        def pp_communicate_encoder_output(encoder_output):
            if self.encoder_runtime_mapping.is_last_pp_rank():
                for pp_rank in self.encoder_runtime_mapping.pp_group:
                    if pp_rank != self.encoder_runtime_mapping.rank:
                        self.nccl_comm.send(encoder_output, pp_rank)
                return encoder_output
            else:
                self.nccl_comm.recv(encoder_output,
                                    self.encoder_runtime_mapping.pp_group[-1])
                return encoder_output

        if self.encoder_runtime_mapping.has_pp():
            # use hidden_states output buffer to receive output as the shapes are same
            encoder_output_buf = outputs[
                'encoder_output'] if self.encoder_runtime_mapping.is_last_pp_rank(
                ) else outputs['hidden_states_output']
            encoder_output = pp_communicate_encoder_output(encoder_output_buf)
        else:
            encoder_output = outputs['encoder_output']

        return encoder_output

    def generate(self,
                 encoder_input_ids,
                 decoder_input_ids,
                 max_new_tokens,
                 num_beams=1,
                 pad_token_id=None,
                 eos_token_id=None,
                 bos_token_id=None,
                 debug_mode=False,
                 return_dict=False,
                 prompt_embedding_table=None,
                 prompt_tasks=None,
                 prompt_vocab_size=None,
                 attention_mask=None,
                 time_encoder=False,
                 return_encoder_output=False):
        ## ensure all externally provided tensors are on the correct device.
        encoder_input_ids = encoder_input_ids.to(self.device)
        decoder_input_ids = decoder_input_ids.to(self.device)

        if attention_mask is not None:
            attention_mask = torch.tensor(attention_mask,
                                          dtype=torch.int32,
                                          device=self.device)

        ## encoder run
        encoder_remove_input_padding = self.encoder_model_config.remove_input_padding if self.encoder_model_config else self.decoder_model_config.remove_input_padding

        encoder_input_ids, encoder_input_lengths, encoder_max_input_length, prompt_tasks = self.process_input(
            encoder_input_ids, encoder_remove_input_padding, pad_token_id,
            prompt_tasks)

        if not self.skip_encoder:
            #logger.info(f"Rank {self.runtime_rank} Running encoder engine ...")
            if time_encoder:
                tik = time.time()
            encoder_output = self.encoder_run(
                encoder_input_ids,
                encoder_input_lengths,
                encoder_max_input_length,
                debug_mode=debug_mode,
                prompt_embedding_table=prompt_embedding_table,
                prompt_tasks=prompt_tasks,
                prompt_vocab_size=prompt_vocab_size,
                attention_mask=attention_mask)
            if time_encoder:
                tok = time.time()
                print(f"TRT-LLM Encoder time {(tok-tik)*1000}ms")
        else:
            encoder_output = prompt_embedding_table
            if encoder_input_ids.dim() > 1:
                encoder_output = encoder_output.unsqueeze(0)

        ## decoder run
        # logger.info(f"Rank {self.runtime_rank} Running decoder engine ...")
        decoder_input_ids, decoder_input_lengths, decoder_max_input_length, _ = self.process_input(
            decoder_input_ids, self.decoder_model_config.remove_input_padding,
            pad_token_id)

        # `cross_attention_mask` in context phase [batch_size, query_len, encoder_input_len]
        # where query_len happens to be 1 in current cases, but not necessarily always, and
        # `cross_attention_mask` in generation phase [batch_size, 1, encoder_input_len] where
        # the query_len is always 1 since we have kv cache.
        cross_attention_mask = None
        if attention_mask is not None:
            cross_attention_mask = torch.tensor(attention_mask,
                                                dtype=torch.int32,
                                                device=self.device).reshape(
                                                    attention_mask.shape[0], 1,
                                                    attention_mask.shape[1])

        # generation config
        sampling_config = SamplingConfig(end_id=eos_token_id,
                                         pad_id=pad_token_id,
                                         num_beams=num_beams,
                                         min_length=1,
                                         return_dict=return_dict)
        sampling_config.update(output_cum_log_probs=return_dict,
                               output_log_probs=return_dict)

        # decoder autoregressive generation
        self.decoder_session.setup(
            decoder_input_lengths.size(0),
            decoder_max_input_length,
            max_new_tokens,
            num_beams,
            max_attention_window_size=None,
            encoder_max_input_length=encoder_max_input_length,
            lora_manager=self.decoder_lora_manager,
            lora_uids=self.lora_task_uids,
        )

        output = self.decoder_session.decode(
            decoder_input_ids,
            decoder_input_lengths,
            sampling_config,
            encoder_output=encoder_output,
            encoder_input_lengths=encoder_input_lengths,
            return_dict=return_dict,
            cross_attention_mask=cross_attention_mask)

        if return_dict and return_encoder_output:
            output['encoder_output'] = encoder_output

        return output



from .extractor import DocExtractor
from .clipper import clip, clip_from_image
from .plot import plot
from .rectangle import Point, Rectangle
from .types import (
  ExtractedResult,
  OCRFragment,
  LayoutClass,
  TableLayoutParsedFormat,
  Layout,
  BaseLayout,
  PlainLayout,
  FormulaLayout,
  TableLayout,
  ModelsDownloader
)



import os
import re

from math import ceil
from PIL.Image import Image
from PIL.ImageOps import expand


def ensure_dir(path: str) -> str:
  path = os.path.abspath(path)
  os.makedirs(path, exist_ok=True)
  return path

def is_space_text(text: str) -> bool:
  return re.match(r"^\s*$", text)

def expand_image(image: Image, percent: float):
  width, height = image.size
  border_width = ceil(width * percent)
  border_height = ceil(height * percent)
  fill_color: tuple[int, ...]

  if image.mode == "RGBA":
    fill_color = (255, 255, 255, 255)
  else:
    fill_color = (255, 255, 255)

  return expand(
    image=image,
    border=(border_width, border_height),
    fill=fill_color,
  )


from typing import Generator
from dataclasses import dataclass
from math import sqrt
from shapely.geometry import Polygon


Point = tuple[float, float]

@dataclass
class Rectangle:
  lt: Point
  rt: Point
  lb: Point
  rb: Point

  def __iter__(self) -> Generator[Point, None, None]:
    yield self.lt
    yield self.lb
    yield self.rb
    yield self.rt

  @property
  def is_valid(self) -> bool:
    return Polygon(self).is_valid

  @property
  def segments(self) -> Generator[tuple[Point, Point], None, None]:
    yield (self.lt, self.lb)
    yield (self.lb, self.rb)
    yield (self.rb, self.rt)
    yield (self.rt, self.lt)

  @property
  def area(self) -> float:
    return Polygon(self).area

  @property
  def size(self) -> tuple[float, float]:
    width: float = 0.0
    height: float = 0.0
    for i, (p1, p2) in enumerate(self.segments):
      dx = p1[0] - p2[0]
      dy = p1[1] - p2[1]
      distance = sqrt(dx * dx + dy * dy)
      if i % 2 == 0:
        height += distance
      else:
        width += distance
    return width / 2, height / 2

  @property
  def wrapper(self) -> tuple[float, float, float, float]:
    x1: float = float("inf")
    y1: float = float("inf")
    x2: float = float("-inf")
    y2: float = float("-inf")
    for x, y in self:
      x1 = min(x1, x)
      y1 = min(y1, y)
      x2 = max(x2, x)
      y2 = max(y2, y)
    return x1, y1, x2, y2

def intersection_area(rect1: Rectangle, rect2: Rectangle) -> float:
  poly1 = Polygon(rect1)
  poly2 = Polygon(rect2)
  if not poly1.is_valid or not poly2.is_valid:
    return 0.0
  intersection = poly1.intersection(poly2)
  if intersection.is_empty:
    return 0.0
  return intersection.area


from typing import Generator
from shapely.geometry import Polygon
from .types import Layout, OCRFragment
from .rectangle import Rectangle


_INCLUDES_MIN_RATE = 0.99

def remove_overlap_layouts(layouts: list[Layout]) -> list[Layout]:
  ctx = _OverlapMatrixContext(layouts)
  # the reason for repeating this multiple times is that deleting a layout
  # may cause its parent layout to change from an originally non-deletable
  # state to a deletable state.
  while True:
    removed_count = len(ctx.removed_indexes)
    for i, layout in enumerate(layouts):
      if i in ctx.removed_indexes or \
         any(0.0 < rate < _INCLUDES_MIN_RATE for rate in ctx.rates_with_other(i)) or \
         all(0.0 == rate for rate in ctx.rates_with_other(i)):
        continue

      if len(layout.fragments) == 0:
        ctx.removed_indexes.add(i)
      else:
        for j in ctx.search_includes_indexes(i):
          ctx.removed_indexes.add(j)
          layout.fragments.extend(layouts[j].fragments)

    if len(ctx.removed_indexes) == removed_count:
      break

  return [
    layout for i, layout in enumerate(layouts)
    if i not in ctx.removed_indexes
  ]

class _OverlapMatrixContext:
  def __init__(self, layouts: list[Layout]):
    length: int = len(layouts)
    polygons: list[Polygon] = [Polygon(layout.rect) for layout in layouts]
    self.rate_matrix: list[list[float]] = [[1.0 for _ in range(length)] for _ in range(length)]
    self.removed_indexes: set[int] = set()
    for i in range(length):
      polygon1 = polygons[i]
      rates = self.rate_matrix[i]
      for j in range(length):
        if i != j:
          polygon2 = polygons[j]
          rates[j] = overlap_rate(polygon1, polygon2)

  def rates_with_other(self, index: int):
    for i, rate in enumerate(self.rate_matrix[index]):
      if i != index and i not in self.removed_indexes:
        yield rate

  def search_includes_indexes(self, index: int):
    for i, rate in enumerate(self.rate_matrix[index]):
      if i != index and \
         i not in self.removed_indexes and \
         rate >= _INCLUDES_MIN_RATE:
        yield i

def merge_fragments_as_line(origin_fragments: list[OCRFragment]) -> list[OCRFragment]:
  fragments: list[OCRFragment] = []
  for group in _split_fragments_into_groups(origin_fragments):
    if len(group) == 1:
      fragments.append(group[0])
      continue

    min_order: float = float("inf")
    texts: list[str] = []
    text_rate_weights: float = 0.0
    proto_texts_len: int = 0

    x1: float = float("inf")
    y1: float = float("inf")
    x2: float = float("-inf")
    y2: float = float("-inf")

    for fragment in sorted(group, key=lambda x: x.rect.lt[0] + x.rect.lb[0]):
      proto_texts_len += len(fragment.text)
      text_rate_weights += fragment.rank * len(fragment.text)
      texts.append(fragment.text)
      min_order = min(min_order, fragment.order)
      for x, y in fragment.rect:
        x1 = min(x1, x)
        y1 = min(y1, y)
        x2 = max(x2, x)
        y2 = max(y2, y)

    if proto_texts_len == 0:
      continue

    fragments.append(OCRFragment(
      order=min_order,
      text=" ".join(texts),
      rank=text_rate_weights / proto_texts_len,
      rect=Rectangle(
        lt=(x1, y1),
        rt=(x2, y1),
        lb=(x1, y2),
        rb=(x2, y2),
      ),
    ))
  return fragments

def _split_fragments_into_groups(fragments: list[OCRFragment]) -> Generator[list[OCRFragment], None, None]:
  group: list[OCRFragment] = []
  sum_height: float = 0.0
  sum_median: float = 0.0
  max_deviation_rate = 0.35

  for fragment in sorted(fragments, key=lambda x: x.rect.lt[1] + x.rect.rt[1]):
    _, y1, _, y2 = fragment.rect.wrapper
    height = y2 - y1
    median = (y1 + y2) / 2.0

    if height == 0:
      continue

    if len(group) > 0:
      next_mean_median = (sum_median + median) / (len(group) + 1)
      next_mean_height = (sum_height + height) / (len(group) + 1)

      deviation_rate = abs(median - next_mean_median) / next_mean_height
      if deviation_rate > max_deviation_rate:
        yield group
        group = []
        sum_height = 0.0
        sum_median = 0.0

    group.append(fragment)
    sum_height += height
    sum_median += median

  if len(group) > 0:
    yield group

# calculating overlap ratio: The reason why area is not used is
# that most of the measurements are of rectangles representing text lines.
# they are very sensitive to changes in height because they are very thin and long.
# In order to make it equally sensitive to length and width, the ratio of area is not used.
def overlap_rate(polygon1: Polygon, polygon2: Polygon) -> float:
  intersection: Polygon = polygon1.intersection(polygon2)
  if intersection.is_empty:
    return 0.0
  else:
    overlay_width, overlay_height = _polygon_size(intersection)
    polygon2_width, polygon2_height = _polygon_size(polygon2)
    if polygon2_width == 0.0 or polygon2_height == 0.0:
      return 0.0
    return (
      overlay_width / polygon2_width +
      overlay_height / polygon2_height
    ) / 2.0

def _polygon_size(polygon: Polygon) -> tuple[float, float]:
  x1: float = float("inf")
  y1: float = float("inf")
  x2: float = float("-inf")
  y2: float = float("-inf")
  for x, y in polygon.exterior.coords:
    x1 = min(x1, x)
    y1 = min(y1, y)
    x2 = max(x2, x)
    y2 = max(y2, y)
  return x2 - x1, y2 - y1


from typing import Iterable
from PIL import ImageDraw
from PIL.ImageFont import load_default, FreeTypeFont
from PIL.Image import Image
from .types import Layout, LayoutClass
from .rectangle import Point

_FRAGMENT_COLOR = (0x49, 0xCF, 0xCB) # Light Green
_Color = tuple[int, int, int]

def plot(image: Image, layouts: Iterable[Layout]) -> None:
  layout_font = load_default(size=35)
  fragment_font = load_default(size=25)
  draw = ImageDraw.Draw(image, mode="RGBA")

  def _draw_number(position: Point, number: int, font: FreeTypeFont, bold: bool, color: _Color) -> None:
    nonlocal draw
    x, y = position
    text = str(object=number)
    width = len(text) * font.size
    offset = round(font.size * 0.15)

    for dx, dy in _generate_delta(bold):
      draw.text(
        xy=(x + dx - width - offset, y + dy),
        text=text,
        font=font,
        fill=color,
      )

  for layout in layouts:
    draw.polygon(
      xy=[p for p in layout.rect],
      outline=_layout_color(layout),
      width=5,
    )

  for layout in layouts:
    for fragment in layout.fragments:
      draw.polygon(
        xy=[p for p in fragment.rect],
        outline=_FRAGMENT_COLOR,
        width=3,
      )
      _draw_number(
        position=fragment.rect.lt,
        number=fragment.order + 1,
        font=fragment_font,
        bold=False,
        color=_FRAGMENT_COLOR,
      )

  for i, layout in enumerate(layouts):
    _draw_number(
      position=layout.rect.lt,
      number=i + 1,
      font=layout_font,
      bold=True,
      color=_layout_color(layout),
    )

def _generate_delta(bold: bool):
  if bold:
    for dx in range(-1, 2):
      for dy in range(-1, 2):
        yield dx, dy
  else:
    yield 0, 0

def _layout_color(layout: Layout) -> _Color:
  cls = layout.cls
  if cls == LayoutClass.TITLE:
    return (0x0A, 0x12, 0x2C) # Dark
  elif cls == LayoutClass.PLAIN_TEXT:
    return (0x3C, 0x67, 0x90) # Blue
  elif cls == LayoutClass.ABANDON:
    return (0xC0, 0xBB, 0xA9) # Gray
  elif cls == LayoutClass.FIGURE:
    return (0x5B, 0x91, 0x3C) # Dark Green
  elif cls == LayoutClass.FIGURE_CAPTION:
    return (0x77, 0xB3, 0x54) # Green
  elif cls == LayoutClass.TABLE:
    return (0x44, 0x17, 0x52) # Dark Purple
  elif cls == LayoutClass.TABLE_CAPTION:
    return (0x81, 0x75, 0xA0) # Purple
  elif cls == LayoutClass.TABLE_FOOTNOTE:
    return (0xEF, 0xB6, 0xC9) # Pink Purple
  elif cls == LayoutClass.ISOLATE_FORMULA:
    return (0xFA, 0x38, 0x27) # Red
  elif cls == LayoutClass.FORMULA_CAPTION:
    return (0xFF, 0x9D, 0x24) # Orange


import onnxruntime

class PredictBase(object):
  def __init__(self):
    pass

  def get_onnx_session(self, model_dir, use_gpu):
    # 使用gpu
    if use_gpu:
      providers = providers=['CUDAExecutionProvider']
    else:
      providers = providers = ['CPUExecutionProvider']

    onnx_session = onnxruntime.InferenceSession(model_dir, None,providers=providers)

    # print("providers:", onnxruntime.get_device())
    return onnx_session


  def get_output_name(self, onnx_session):
    """
    output_name = onnx_session.get_outputs()[0].name
    :param onnx_session:
    :return:
    """
    output_name = []
    for node in onnx_session.get_outputs():
      output_name.append(node.name)
    return output_name

  def get_input_name(self, onnx_session):
    """
    input_name = onnx_session.get_inputs()[0].name
    :param onnx_session:
    :return:
    """
    input_name = []
    for node in onnx_session.get_inputs():
      input_name.append(node.name)
    return input_name

  def get_input_feed(self, input_name, image_numpy):
    """
    input_feed={self.input_name: image_numpy}
    :param input_name:
    :param image_numpy:
    :return:
    """
    input_feed = {}
    for name in input_name:
      input_feed[name] = image_numpy
    return input_feed


from .operators import *


def transform(data, ops=None):
  """transform"""
  if ops is None:
    ops = []
  for op in ops:
    data = op(data)
    if data is None:
      return None
  return data


def create_operators(op_param_list, global_config=None):
  """
  create operators based on the config

  Args:
    params(list): a dict list, used to create some operators
  """
  assert isinstance(op_param_list, list), "operator config should be a list"
  ops = []
  for operator in op_param_list:
    assert isinstance(operator, dict) and len(operator) == 1, "yaml format error"
    op_name = list(operator)[0]
    param = {} if operator[op_name] is None else operator[op_name]
    if global_config is not None:
      param.update(global_config)
    op = eval(op_name)(**param)
    ops.append(op)
  return ops




from typing import Literal, Generator
from PIL.Image import Image
from doclayout_yolo import YOLOv10
from logging import Logger, getLogger

from .models import HuggingfaceModelsDownloader
from .ocr import OCR
from .ocr_corrector import correct_fragments
from .raw_optimizer import RawOptimizer
from .rectangle import intersection_area, Rectangle
from .table import Table
from .latex import LaTeX
from .layout_order import LayoutOrder
from .overlap import merge_fragments_as_line, remove_overlap_layouts
from .clipper import clip_from_image
from .types import (
  ExtractedResult,
  ModelsDownloader,
  OCRFragment,
  TableLayoutParsedFormat,
  Layout,
  LayoutClass,
  PlainLayout,
  TableLayout,
  FormulaLayout,
)


class DocExtractor:
  def __init__(
      self,
      model_cache_dir: str | None = None,
      device: Literal["cpu", "cuda"] = "cpu",
      ocr_for_each_layouts: bool = True,
      extract_formula: bool = True,
      extract_table_format: TableLayoutParsedFormat | None = None,
      models_downloader: ModelsDownloader | None = None,
      logger: Logger | None = None,
    ):
    self._logger = logger or getLogger(__name__)
    logger.info("DocExtractor test initialized.")
    self._models_downloader = models_downloader or HuggingfaceModelsDownloader(self._logger, model_cache_dir)

    self._device: Literal["cpu", "cuda"] = device
    self._ocr_for_each_layouts: bool = ocr_for_each_layouts
    self._extract_formula: bool = extract_formula
    self._extract_table_format: TableLayoutParsedFormat | None = extract_table_format
    self._yolo: YOLOv10 | None = None
    self._ocr: OCR = OCR(
      device=device,
      get_model_dir=self._models_downloader.onnx_ocr,
    )
    self._table: Table = Table(
      device=device,
      get_model_dir=self._models_downloader.struct_eqtable,
    )
    self._latex: LaTeX = LaTeX(
      get_model_dir=self._models_downloader.latex,
      device=device,
    )
    self._layout_order: LayoutOrder = LayoutOrder(
      get_model_dir=self._models_downloader.layoutreader,
    )

  def extract(
      self,
      image: Image,
      adjust_points: bool = False,
    ) -> ExtractedResult:

    raw_optimizer = RawOptimizer(image, adjust_points)
    fragments = list(self._ocr.search_fragments(raw_optimizer.image_np))
    raw_optimizer.receive_raw_fragments(fragments)
    layouts = list(self._yolo_extract_layouts(raw_optimizer.image))
    layouts = self._layouts_matched_by_fragments(fragments, layouts)
    layouts = remove_overlap_layouts(layouts)

    if self._ocr_for_each_layouts:
      self._correct_fragments_by_ocr_layouts(raw_optimizer.image, layouts)

    layouts = self._layout_order.sort(layouts, raw_optimizer.image.size)
    layouts = [layout for layout in layouts if self._should_keep_layout(layout)]

    self._parse_table_and_formula_layouts(layouts, raw_optimizer)

    for layout in layouts:
      layout.fragments = merge_fragments_as_line(layout.fragments)

    raw_optimizer.receive_raw_layouts(layouts)

    return ExtractedResult(
      rotation=raw_optimizer.rotation,
      layouts=layouts,
      extracted_image=image,
      adjusted_image=raw_optimizer.adjusted_image,
    )

  def _yolo_extract_layouts(self, source: Image) -> Generator[Layout, None, None]:
    # about source parameter to see:
    # https://github.com/opendatalab/DocLayout-YOLO/blob/7c4be36bc61f11b67cf4a44ee47f3c41e9800a91/doclayout_yolo/data/build.py#L157-L175
    det_res = self._get_yolo().predict(
      source=source,
      imgsz=1024,
      conf=0.2,
      device=self._device    # Device to use (e.g., "cuda" or "cpu")
    )
    boxes = det_res[0].__dict__["boxes"]

    for cls_id, rect in zip(boxes.cls, boxes.xyxy):
      cls_id = cls_id.item()
      cls=LayoutClass(round(cls_id))

      x1, y1, x2, y2 = rect
      x1 = x1.item()
      y1 = y1.item()
      x2 = x2.item()
      y2 = y2.item()
      rect = Rectangle(
        lt=(x1, y1),
        rt=(x2, y1),
        lb=(x1, y2),
        rb=(x2, y2),
      )
      if rect.is_valid:
        if cls == LayoutClass.TABLE:
          yield TableLayout(cls=cls, rect=rect, fragments=[], parsed=None)
        elif cls == LayoutClass.ISOLATE_FORMULA:
          yield FormulaLayout(cls=cls, rect=rect, fragments=[], latex=None)
        else:
          yield PlainLayout(cls=cls, rect=rect, fragments=[])

  def _layouts_matched_by_fragments(self, fragments: list[OCRFragment], layouts: list[Layout]):
    layouts_group = self._split_layouts_by_group(layouts)
    for fragment in fragments:
      for sub_layouts in layouts_group:
        layout = self._find_matched_layout(fragment, sub_layouts)
        if layout is not None:
          layout.fragments.append(fragment)
          break
    return layouts

  def _correct_fragments_by_ocr_layouts(self, source: Image, layouts: list[Layout]):
    for layout in layouts:
      correct_fragments(self._ocr, source, layout)

  def _parse_table_and_formula_layouts(self, layouts: list[Layout], raw_optimizer: RawOptimizer):
    for layout in layouts:
      if isinstance(layout, FormulaLayout) and self._extract_formula:
        image = clip_from_image(raw_optimizer.image, layout.rect)
        layout.latex = self._latex.extract(image)
      elif isinstance(layout, TableLayout) and self._extract_table_format is not None:
        image = clip_from_image(raw_optimizer.image, layout.rect)
        parsed = self._table.predict(image, self._extract_table_format)
        if parsed is not None:
          layout.parsed = (parsed, self._extract_table_format)

  def _split_layouts_by_group(self, layouts: list[Layout]):
    texts_layouts: list[Layout] = []
    abandon_layouts: list[Layout] = []

    for layout in layouts:
      cls = layout.cls
      if cls == LayoutClass.TITLE or \
         cls == LayoutClass.PLAIN_TEXT or \
         cls == LayoutClass.FIGURE_CAPTION or \
         cls == LayoutClass.TABLE_CAPTION or \
         cls == LayoutClass.TABLE_FOOTNOTE or \
         cls == LayoutClass.FORMULA_CAPTION:
        texts_layouts.append(layout)
      elif cls == LayoutClass.ABANDON:
        abandon_layouts.append(layout)

    return texts_layouts, abandon_layouts

  def _find_matched_layout(self, fragment: OCRFragment, layouts: list[Layout]) -> Layout | None:
    fragment_area = fragment.rect.area
    primary_layouts: list[(Layout, float)] = []

    if fragment_area == 0.0:
      return None

    for layout in layouts:
      area = intersection_area(fragment.rect, layout.rect)
      if area / fragment_area > 0.85:
        primary_layouts.append((layout, layout.rect.area))

    min_area: float = float("inf")
    min_layout: Layout | None = None

    for layout, area in primary_layouts:
      if area < min_area:
        min_area = area
        min_layout = layout

    return min_layout

  def _get_yolo(self) -> YOLOv10:
    if self._yolo is None:
      model_path = self._models_downloader.yolo()
      self._yolo = YOLOv10(str(model_path))
    return self._yolo

  def _should_keep_layout(self, layout: Layout) -> bool:
    if len(layout.fragments) > 0:
      return True
    cls = layout.cls
    return (
      cls == LayoutClass.FIGURE or
      cls == LayoutClass.TABLE or
      cls == LayoutClass.ISOLATE_FORMULA
    )




import numpy as np
import cv2

def get_rotate_crop_image(img, points):
  """
  img_height, img_width = img.shape[0:2]
  left = int(np.min(points[:, 0]))
  right = int(np.max(points[:, 0]))
  top = int(np.min(points[:, 1]))
  bottom = int(np.max(points[:, 1]))
  img_crop = img[top:bottom, left:right, :].copy()
  points[:, 0] = points[:, 0] - left
  points[:, 1] = points[:, 1] - top
  """
  assert len(points) == 4, "shape of points must be 4*2"
  img_crop_width = int(
    max(
      np.linalg.norm(points[0] - points[1]), np.linalg.norm(points[2] - points[3])
    )
  )
  img_crop_height = int(
    max(
      np.linalg.norm(points[0] - points[3]), np.linalg.norm(points[1] - points[2])
    )
  )
  pts_std = np.float32(
    [
      [0, 0],
      [img_crop_width, 0],
      [img_crop_width, img_crop_height],
      [0, img_crop_height],
    ]
  )
  M = cv2.getPerspectiveTransform(points, pts_std)
  dst_img = cv2.warpPerspective(
    img,
    M,
    (img_crop_width, img_crop_height),
    borderMode=cv2.BORDER_REPLICATE,
    flags=cv2.INTER_CUBIC,
  )
  dst_img_height, dst_img_width = dst_img.shape[0:2]
  if dst_img_height * 1.0 / dst_img_width >= 1.5:
    dst_img = np.rot90(dst_img)
  return dst_img


def get_minarea_rect_crop(img, points):
  bounding_box = cv2.minAreaRect(np.array(points).astype(np.int32))
  points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

  index_a, index_b, index_c, index_d = 0, 1, 2, 3
  if points[1][1] > points[0][1]:
    index_a = 0
    index_d = 1
  else:
    index_a = 1
    index_d = 0
  if points[3][1] > points[2][1]:
    index_b = 2
    index_c = 3
  else:
    index_b = 3
    index_c = 2

  box = [points[index_a], points[index_b], points[index_c], points[index_d]]
  crop_img = get_rotate_crop_image(img, np.array(box))
  return crop_img

def str2bool(v):
  return v.lower() in ("true", "t", "1")


import os
import unittest

from PIL import Image
from doc_page_extractor import DocExtractor, Layout, LayoutClass


class TestGroup(unittest.TestCase):
  def test_history_bugs(self):
    model_path = os.path.join(self._project_path(), "model")
    image_path = os.path.join(self._project_path(), "tests", "images", "figure.png")
    os.makedirs(model_path, exist_ok=True)

    extractor = DocExtractor(model_path, "cpu")
    layouts: list[tuple[LayoutClass, list[str]]]

    with Image.open(image_path) as image:
      result = extractor.extract(image, "ch")
      layouts = [self._format_Layout(layout) for layout in result.layouts]

    self.assertEqual(layouts, [
      (LayoutClass.PLAIN_TEXT, [
        "口的11.8%①。这既是江南农业落后的反映，又是它的原因。当战国以",
        "后黄河流域因铁器牛耕的普及获得基本的开发，农区联结成一大片的",
        "时候，南方农业开发始终没有突破星点状或斑块状分布的格局。由于",
        "地旷人稀，耕作相当粗放，许多水田采取火耕水瓣的方式，旱田则多",
        "行刀耕火种②。司马迁在《史记·货殖列传》中说：“总之，楚越之",
        "地，地厂人希，饭稻囊鱼，或火耕而水瓣，果隋（蕨）赢（螺）蛤，",
        "不待贾而足，地势饶食，无饥谨之患，以故皆偷生，无积聚而多",
        "贫。”这种概括虽然未免太突出了南方经济的落后面，有一定片面性，",
        "但大体还是反映了实际情形的。战国秦汉时期，南方与黄河流域农业",
        "的差距显然拉大了。",
      ]),
      (LayoutClass.FIGURE, []),
      (LayoutClass.FIGURE_CAPTION, [
        "西晋陶水田犁耙模型（广东连县出土）"
      ]),
      (LayoutClass.FIGURE, []),
      (LayoutClass.FIGURE_CAPTION, [
        "南朝陶耙田模型 （广西苍梧倒水出土）"
      ]),
      (LayoutClass.PLAIN_TEXT, [
        "①据赵文林、谢淑君：《中国人口史》（人民出版社1988年）有关资料统计。",
        "②《盐铁论·通有》：“荆扬…………伐木而树谷，焚莱而播粟，火耕而水。”"
      ]),
      (LayoutClass.ABANDON, [
        "136"
      ]),
    ])

  def _format_Layout(self, layout: Layout) -> tuple[LayoutClass, list[str]]:
    return layout.cls, [f.text.strip() for f in layout.fragments]

  def _project_path(self) -> str:
    return os.path.abspath(os.path.join(__file__, "..", ".."))


# Copy from https://github.com/ppaanngggg/layoutreader/blob/main/v3/helpers.py
from collections import defaultdict
from typing import List, Dict

import torch
from transformers import LayoutLMv3ForTokenClassification

MAX_LEN = 510
CLS_TOKEN_ID = 0
UNK_TOKEN_ID = 3
EOS_TOKEN_ID = 2


class DataCollator:
  def __call__(self, features: List[dict]) -> Dict[str, torch.Tensor]:
    bbox = []
    labels = []
    input_ids = []
    attention_mask = []

    # clip bbox and labels to max length, build input_ids and attention_mask
    for feature in features:
      _bbox = feature["source_boxes"]
      if len(_bbox) > MAX_LEN:
        _bbox = _bbox[:MAX_LEN]
      _labels = feature["target_index"]
      if len(_labels) > MAX_LEN:
        _labels = _labels[:MAX_LEN]
      _input_ids = [UNK_TOKEN_ID] * len(_bbox)
      _attention_mask = [1] * len(_bbox)
      assert len(_bbox) == len(_labels) == len(_input_ids) == len(_attention_mask)
      bbox.append(_bbox)
      labels.append(_labels)
      input_ids.append(_input_ids)
      attention_mask.append(_attention_mask)

    # add CLS and EOS tokens
    for i in range(len(bbox)):
      bbox[i] = [[0, 0, 0, 0]] + bbox[i] + [[0, 0, 0, 0]]
      labels[i] = [-100] + labels[i] + [-100]
      input_ids[i] = [CLS_TOKEN_ID] + input_ids[i] + [EOS_TOKEN_ID]
      attention_mask[i] = [1] + attention_mask[i] + [1]

    # padding to max length
    max_len = max(len(x) for x in bbox)
    for i in range(len(bbox)):
      bbox[i] = bbox[i] + [[0, 0, 0, 0]] * (max_len - len(bbox[i]))
      labels[i] = labels[i] + [-100] * (max_len - len(labels[i]))
      input_ids[i] = input_ids[i] + [EOS_TOKEN_ID] * (max_len - len(input_ids[i]))
      attention_mask[i] = attention_mask[i] + [0] * (
        max_len - len(attention_mask[i])
      )

    ret = {
      "bbox": torch.tensor(bbox),
      "attention_mask": torch.tensor(attention_mask),
      "labels": torch.tensor(labels),
      "input_ids": torch.tensor(input_ids),
    }
    # set label > MAX_LEN to -100, because original labels may be > MAX_LEN
    ret["labels"][ret["labels"] > MAX_LEN] = -100
    # set label > 0 to label-1, because original labels are 1-indexed
    ret["labels"][ret["labels"] > 0] -= 1
    return ret


def boxes2inputs(boxes: List[List[int]]) -> Dict[str, torch.Tensor]:
  bbox = [[0, 0, 0, 0]] + boxes + [[0, 0, 0, 0]]
  input_ids = [CLS_TOKEN_ID] + [UNK_TOKEN_ID] * len(boxes) + [EOS_TOKEN_ID]
  attention_mask = [1] + [1] * len(boxes) + [1]
  return {
    "bbox": torch.tensor([bbox]),
    "attention_mask": torch.tensor([attention_mask]),
    "input_ids": torch.tensor([input_ids]),
  }


def prepare_inputs(
  inputs: Dict[str, torch.Tensor], model: LayoutLMv3ForTokenClassification
) -> Dict[str, torch.Tensor]:
  ret = {}
  for k, v in inputs.items():
    v = v.to(model.device)
    if torch.is_floating_point(v):
      v = v.to(model.dtype)
    ret[k] = v
  return ret


def parse_logits(logits: torch.Tensor, length: int) -> List[int]:
  """
  parse logits to orders

  :param logits: logits from model
  :param length: input length
  :return: orders
  """
  logits = logits[1 : length + 1, :length]
  orders = logits.argsort(descending=False).tolist()
  ret = [o.pop() for o in orders]
  while True:
    order_to_idxes = defaultdict(list)
    for idx, order in enumerate(ret):
      order_to_idxes[order].append(idx)
    # filter idxes len > 1
    order_to_idxes = {k: v for k, v in order_to_idxes.items() if len(v) > 1}
    if not order_to_idxes:
      break
    # filter
    for order, idxes in order_to_idxes.items():
      # find original logits of idxes
      idxes_to_logit = {}
      for idx in idxes:
        idxes_to_logit[idx] = logits[idx, order]
      idxes_to_logit = sorted(
        idxes_to_logit.items(), key=lambda x: x[1], reverse=True
      )
      # keep the highest logit as order, set others to next candidate
      for idx, _ in idxes_to_logit[1:]:
        ret[idx] = orders[idx].pop()

  return ret


def check_duplicate(a: List[int]) -> bool:
  return len(a) != len(set(a))
